import colorama as cra
from .common import *
from .math import *
from .geometry import *
from .scene import *
from .frame import *
from .collision import *
from .dynamics import *
from .render import *

cra.init()



from typing import List

import numpy as np
import taichi as ti

from .common.config import Config
from .common.ti_viewport import Viewport
from .dynamics.ti_phy_world import PhysicsWorld
from .frame import Frame


@ti.data_oriented
class Scene():
    def __init__(self, name: str, width: int = 1280, height: int = 720):
        self._gui = ti.GUI(name,
                           res=(width, height),
                           background_color=Config.BackgroundColor)

        # physics world
        self._world = PhysicsWorld()
        # calcuate step
        self._fps = 120
        self._dt = 1 / self._fps
        # viewport system
        self._viewport = Viewport(ti.Vector([0, height]), ti.Vector([width,
                                                                     0]))
        self._origin: ti.Vector = ti.Vector(
            [self._viewport.width / 2.0, self._viewport.height / 2.0])
        self._xform: ti.Vector = ti.Vector([0.0, 0.0])
        # zoom system
        self._meter_to_pixel: float = 33.0
        self._pixel_to_meter: float = 1 / self._meter_to_pixel
        self._target_meter_to_pixel: float = 53.0
        self._target_pixel_to_meter: float = 1 / self._target_meter_to_pixel
        self._restit: float = 2.0
        self._delta_time: float = 15.0
        # axis system
        self._axis_p_cnt: int = 10
        self._axis_p_list = ti.Vector.field(2,
                                            float,
                                            shape=4 * self._axis_p_cnt)
        self._axis_lin_st = ti.Vector.field(2, float, shape=2)
        self._axis_lin_ed = ti.Vector.field(2, float, shape=2)

        # mouse
        # NOTE: some algorithm need to cacluate the pos's len
        # in init state
        self._mouse_pos: ti.Vector = ti.Vector([1.0, 1.0])
        # the right-mouse btn drag move flag(change viewport)
        self._mouse_viewport_move: bool = False

        # extern frame table
        self._ext_frame_list: List[Frame] = []
        self._ext_frame_idx: int = 0

    def physics_sim(self) -> None:
        self._world.step_velocity(self._dt)
        self._world.step_position(self._dt)

    def render(self) -> None:
        self.smooth_scale()
        # self.render_axis()
        self.render_body()
        self.render_center()
        self.render_rot_line()

    def register_frame(self, frame: Frame) -> None:
        self._ext_frame_list.append(frame)

    def remove_frame(self, frame: Frame) -> None:
        # NOTE: need to check if exist first
        self._ext_frame_list.remove(frame)

    def clear_all(self) -> None:
        self._world.clear_all_bodies()
        self._world.clear_all_joints()
        # self._maintainer.clear_all()
        # self._dbvt.clear_all()

    def calc_nxt_frame(self, delta: int) -> None:
        ext_len: int = len(self._ext_frame_list)
        assert -ext_len <= delta <= ext_len

        self._ext_frame_idx = Config.clamp(self._ext_frame_idx, 0, ext_len - 1)
        # NOTE: the sign of mod in python depend on dividend
        self._ext_frame_idx = (self._ext_frame_idx + delta) % ext_len

    def init_frame(self) -> None:
        self.change_frame(0)
        # set the data to the world
        self._world.init_data()

    def change_frame(self, delta: int) -> None:
        self.clear_all()
        self.calc_nxt_frame(delta)
        self._ext_frame_list[self._ext_frame_idx].load()

    @ti.func
    def world_to_screen(self, world, scale, origx, origy, xformx, xformy, vw,
                        vh):
        origin = ti.Vector([origx + xformx, origy + xformy])

        resx = (origin.x + world.x * scale) / vw
        resy = (origin.y + world.y * scale) / vh
        return ti.Vector([resx, resy])

    def screen_to_world(self, pos: ti.Vector) -> ti.Vector:
        orign: ti.Vector = ti.Vector(
            [self._origin.x + self._xform.x, self._origin.y + self._xform.y])

        vw: float = self._viewport.width
        vh: float = self._viewport.height

        res: ti.Vector = ti.Vector([0.0, 0.0])
        res.x = pos.x * vw
        res.y = pos.y * vh
        res = res - orign
        res = res * self._pixel_to_meter

        return res

    @ti.kernel
    def gen_axis_data(self, scale: float, origx: float, origy: float,
                      xformx: float, xformy: float, vw: float, vh: float):
        tmp = ti.Vector([0.0, 0.0])
        for i in range(-self._axis_p_cnt, self._axis_p_cnt + 1):
            tmp = self.world_to_screen(ti.Vector([i * 1.0, 0.0]), scale, origx,
                                       origy, xformx, xformy, vw, vh)
            self._axis_p_list[(i + self._axis_p_cnt) * 2] = tmp

            tmp = self.world_to_screen(ti.Vector([0.0, i * 1.0]), scale, origx,
                                       origy, xformx, xformy, vw, vh)
            self._axis_p_list[(i + self._axis_p_cnt) * 2 + 1] = tmp

        self._axis_lin_st[0] = self._axis_p_list[0]
        self._axis_lin_st[1] = self._axis_p_list[1]
        self._axis_lin_ed[0] = self._axis_p_list[4 * self._axis_p_cnt - 2]
        self._axis_lin_ed[1] = self._axis_p_list[4 * self._axis_p_cnt - 1]

    def render_axis(self) -> None:
        self.gen_axis_data(self._meter_to_pixel, self._origin.x,
                           self._origin.y, self._xform.x, self._xform.y,
                           self._viewport.width, self._viewport.height)

        self._gui.circles(self._axis_p_list.to_numpy(),
                          radius=2,
                          color=Config.AxisPointColor)

        self._gui.lines(begin=self._axis_lin_st.to_numpy(),
                        end=self._axis_lin_ed.to_numpy(),
                        color=Config.AxisLineColor)

    @ti.func
    def rotate_mat(self, radian):
        cos_val = ti.cos(radian)
        sin_val = ti.sin(radian)
        res = ti.Matrix([[cos_val, -sin_val], [sin_val, cos_val]])

        return res

    @ti.kernel
    def gen_body_data(self, scale: float, origx: float, origy: float,
                      xformx: float, xformy: float, vw: float, vh: float):
        for i in range(len(self._world._body_list)):
            rot_mat = self.rotate_mat(self._world._rot[i])

            if self._world._shape_type[i] == 1:
                self._world._cirspos[i] = self.world_to_screen(
                    self._world._cirpos[i], scale, origx, origy, xformx,
                    xformy, vw, vh)
                self._world._scir_rad[i] = self._world._cir_rad[i] * scale

            elif self._world._shape_type[i] == 2:
                pass
            elif self._world._shape_type[i] == 3:
                self._world._poly_trispos[i] = self.world_to_screen(
                    self._world._poly_tripos[i], scale, origx, origy, xformx,
                    xformy, vw, vh)
                for j in range(3):
                    self._world._poly_trisst[i, j] = self.world_to_screen(
                        self._world._poly_tripos[i] +
                        rot_mat @ self._world._poly_trist[i, j], scale, origx, origy,
                        xformx, xformy, vw, vh)

                    self._world._poly_trised[i, j] = self.world_to_screen(
                        self._world._poly_tripos[i] +
                        rot_mat @ self._world._poly_tried[i, j], scale, origx, origy,
                        xformx, xformy, vw, vh)

                self._world._poly_tri_a[i, 0] = self._world._poly_trisst[i, 0]
                self._world._poly_tri_b[i, 0] = self._world._poly_trisst[i, 1]
                self._world._poly_tri_c[i, 0] = self._world._poly_trisst[i, 2]

            elif self._world._shape_type[i] == 4:
                self._world._poly_recspos[i] = self.world_to_screen(
                    self._world._poly_recpos[i], scale, origx, origy, xformx,
                    xformy, vw, vh)
                for j in range(4):
                    self._world._poly_recsst[i, j] = self.world_to_screen(
                        self._world._poly_recpos[i] +
                        rot_mat @ self._world._poly_recst[i, j], scale, origx, origy,
                        xformx, xformy, vw, vh)

                    self._world._poly_recsed[i, j] = self.world_to_screen(
                        self._world._poly_recpos[i] +
                        rot_mat @ self._world._poly_reced[i, j], scale, origx, origy,
                        xformx, xformy, vw, vh)

                    if j >= 2:
                        self._world._poly_rec_a[
                            i, j - 2] = self._world._poly_recsst[i, 0]
                        self._world._poly_rec_b[
                            i, j - 2] = self._world._poly_recsst[i, j - 1]
                        self._world._poly_rec_c[
                            i, j - 2] = self._world._poly_recsst[i, j]

            elif self._world._shape_type[i] == 5:
                self._world._poly_penspos[i] = self.world_to_screen(
                    self._world._poly_penpos[i], scale, origx, origy, xformx,
                    xformy, vw, vh)
                for j in range(5):
                    self._world._poly_pensst[i, j] = self.world_to_screen(
                        self._world._poly_penpos[i] +
                        rot_mat @ self._world._poly_penst[i, j], scale, origx, origy,
                        xformx, xformy, vw, vh)

                    self._world._poly_pensed[i, j] = self.world_to_screen(
                        self._world._poly_penpos[i] +
                        rot_mat @ self._world._poly_pened[i, j], scale, origx, origy,
                        xformx, xformy, vw, vh)

                    if j >= 2:
                        self._world._poly_pen_a[
                            i, j - 2] = self._world._poly_pensst[i, 0]
                        self._world._poly_pen_b[
                            i, j - 2] = self._world._poly_pensst[i, j - 1]
                        self._world._poly_pen_c[
                            i, j - 2] = self._world._poly_pensst[i, j]

            elif self._world._shape_type[i] == 6:
                pass

            elif self._world._shape_type[i] == 7:
                self._world._cap_spos[i] = self.world_to_screen(
                    self._world._cap_pos[i], scale, origx, origy, xformx,
                    xformy, vw, vh)

                tmp1 = ti.Vector([0.0, 0.0])
                tmp2 = ti.Vector([0.0, 0.0])
                offset = ti.min(self._world._cap_width[i],
                                self._world._cap_height[i]) / 2.0
                if self._world._cap_width[i] > self._world._cap_height[i]:
                    tmp1.x = -(self._world._cap_width[i] / 2.0 - offset)
                    tmp2.x = (self._world._cap_width[i] / 2.0 - offset)
                else:
                    tmp1.y = -(self._world._cap_height[i] / 2.0 - offset)
                    tmp2.y = (self._world._cap_height[i] / 2.0 - offset)

                self._world._cap_c1[i] = self.world_to_screen(
                    self._world._cap_pos[i] + rot_mat @ tmp1, scale, origx, origy,
                    xformx, xformy, vw, vh)

                self._world._cap_c2[i] = self.world_to_screen(
                    self._world._cap_pos[i] + rot_mat @ tmp2, scale, origx, origy,
                    xformx, xformy, vw, vh)

                self._world._cap_p[i, 0] = ti.Vector([
                    -self._world._cap_width[i] / 2.0,
                    -self._world._cap_height[i] / 2.0
                ])
                self._world._cap_p[i, 1] = ti.Vector([
                    self._world._cap_width[i] / 2.0,
                    -self._world._cap_height[i] / 2.0
                ])
                self._world._cap_p[i, 2] = ti.Vector([
                    self._world._cap_width[i] / 2.0,
                    self._world._cap_height[i] / 2.0
                ])
                self._world._cap_p[i, 3] = ti.Vector([
                    -self._world._cap_width[i] / 2.0,
                    self._world._cap_height[i] / 2.0
                ])

                if self._world._cap_width[i] > self._world._cap_height[i]:
                    self._world._cap_p[i, 0].x += offset
                    self._world._cap_p[i, 1].x -= offset
                    self._world._cap_p[i, 2].x -= offset
                    self._world._cap_p[i, 3].x += offset
                else:
                    self._world._cap_p[i, 0].y += offset
                    self._world._cap_p[i, 1].y += offset
                    self._world._cap_p[i, 2].y -= offset
                    self._world._cap_p[i, 3].y -= offset

                for j in range(2):
                    self._world._cap_rec_a[i, j] = self.world_to_screen(
                        self._world._cap_pos[i] + rot_mat @ self._world._cap_p[i, 0],
                        scale, origx, origy, xformx, xformy, vw, vh)

                    self._world._cap_rec_b[i, j] = self.world_to_screen(
                        self._world._cap_pos[i] + rot_mat @ self._world._cap_p[i, j + 1],
                        scale, origx, origy, xformx, xformy, vw, vh)

                    self._world._cap_rec_c[i, j] = self.world_to_screen(
                        self._world._cap_pos[i] + rot_mat @ self._world._cap_p[i, j + 2],
                        scale, origx, origy, xformx, xformy, vw, vh)

    def render_body(self) -> None:
        # self._world.random_set()
        self.gen_body_data(self._meter_to_pixel, self._origin.x,
                           self._origin.y, self._xform.x, self._xform.y,
                           self._viewport.width, self._viewport.height)

        self._gui.circles(self._world._cirspos.to_numpy(),
                          radius=self._world._scir_rad[1],
                          color=Config.FillColor)

        tri_st = self._world._poly_trisst.to_numpy()
        tri_ed = self._world._poly_trised.to_numpy()
        tri_st = tri_st.reshape(tri_st.shape[0] * tri_st.shape[1], -1)
        tri_ed = tri_ed.reshape(tri_ed.shape[0] * tri_ed.shape[1], -1)
        self._gui.lines(begin=tri_st,
                        end=tri_ed,
                        color=Config.OuterLineColor,
                        radius=2)

        rec_st = self._world._poly_recsst.to_numpy()
        rec_ed = self._world._poly_recsed.to_numpy()
        rec_st = rec_st.reshape(rec_st.shape[0] * rec_st.shape[1], -1)
        rec_ed = rec_ed.reshape(rec_ed.shape[0] * rec_ed.shape[1], -1)
        self._gui.lines(begin=rec_st,
                        end=rec_ed,
                        color=Config.OuterLineColor,
                        radius=2)

        pen_st = self._world._poly_pensst.to_numpy()
        pen_ed = self._world._poly_pensed.to_numpy()
        pen_st = pen_st.reshape(pen_st.shape[0] * pen_st.shape[1], -1)
        pen_ed = pen_ed.reshape(pen_ed.shape[0] * pen_ed.shape[1], -1)
        self._gui.lines(begin=pen_st,
                        end=pen_ed,
                        color=Config.OuterLineColor,
                        radius=2)

        tri_a = self._world._poly_tri_a.to_numpy()
        tri_b = self._world._poly_tri_b.to_numpy()
        tri_c = self._world._poly_tri_c.to_numpy()
        tri_a = tri_a.reshape(tri_a.shape[0] * tri_a.shape[1], -1)
        tri_b = tri_b.reshape(tri_b.shape[0] * tri_b.shape[1], -1)
        tri_c = tri_c.reshape(tri_c.shape[0] * tri_c.shape[1], -1)
        self._gui.triangles(a=tri_a, b=tri_b, c=tri_c, color=Config.FillColor)

        rec_a = self._world._poly_rec_a.to_numpy()
        rec_b = self._world._poly_rec_b.to_numpy()
        rec_c = self._world._poly_rec_c.to_numpy()
        rec_a = rec_a.reshape(rec_a.shape[0] * rec_a.shape[1], -1)
        rec_b = rec_b.reshape(rec_b.shape[0] * rec_b.shape[1], -1)
        rec_c = rec_c.reshape(rec_c.shape[0] * rec_c.shape[1], -1)
        self._gui.triangles(a=rec_a, b=rec_b, c=rec_c, color=Config.FillColor)

        pen_a = self._world._poly_pen_a.to_numpy()
        pen_b = self._world._poly_pen_b.to_numpy()
        pen_c = self._world._poly_pen_c.to_numpy()
        pen_a = pen_a.reshape(pen_a.shape[0] * pen_a.shape[1], -1)
        pen_b = pen_b.reshape(pen_b.shape[0] * pen_b.shape[1], -1)
        pen_c = pen_c.reshape(pen_c.shape[0] * pen_c.shape[1], -1)
        self._gui.triangles(a=pen_a, b=pen_b, c=pen_c, color=Config.FillColor)

        offset = np.fmin(self._world._cap_width[0],
                         self._world._cap_height[0]) / 2.0
        self._gui.circles(self._world._cap_c1.to_numpy(),
                          radius=self._meter_to_pixel * offset,
                          color=Config.FillColor)

        self._gui.circles(self._world._cap_c2.to_numpy(),
                          radius=self._meter_to_pixel * offset,
                          color=Config.FillColor)

        rec_a = self._world._cap_rec_a.to_numpy()
        rec_b = self._world._cap_rec_b.to_numpy()
        rec_c = self._world._cap_rec_c.to_numpy()
        rec_a = rec_a.reshape(rec_a.shape[0] * rec_a.shape[1], -1)
        rec_b = rec_b.reshape(rec_b.shape[0] * rec_b.shape[1], -1)
        rec_c = rec_c.reshape(rec_c.shape[0] * rec_c.shape[1], -1)
        self._gui.triangles(a=rec_a, b=rec_b, c=rec_c, color=Config.FillColor)

    def render_center(self) -> None:
        self._gui.circles(self._world._cirspos.to_numpy(),
                          color=Config.BodyCenterColor,
                          radius=4)

        self._gui.circles(self._world._poly_trispos.to_numpy(),
                          color=Config.BodyCenterColor,
                          radius=4)

        self._gui.circles(self._world._poly_recspos.to_numpy(),
                          color=Config.BodyCenterColor,
                          radius=4)

        self._gui.circles(self._world._poly_penspos.to_numpy(),
                          color=Config.BodyCenterColor,
                          radius=4)

        self._gui.circles(self._world._cap_spos.to_numpy(),
                          color=Config.BodyCenterColor,
                          radius=4)

    def render_rot_line(self) -> None:
        xpos = ti.Vector([0.12, 0.0])
        ypos = ti.Vector([0.12, 0.0])
        

    def handle_mouse_move_evt(self, x: float, y: float) -> None:
        cur_pos: ti.Vector = self.screen_to_world(ti.Vector([x, y]))
        delta_pos: ti.Vector = cur_pos - self._mouse_pos

        if self._mouse_viewport_move:
            delta_pos = delta_pos * self._meter_to_pixel
            # 0.5 just a hack value for equal radio move offset
            # 33.0 is init meter_to_pixel value
            self._xform = self._xform + delta_pos * 0.5 * (
                33.0 / self._meter_to_pixel)
        self._mouse_pos = cur_pos

        return

    def handle_right_mouse_evt(self, state) -> None:
        if state == ti.GUI.PRESS:
            self._mouse_viewport_move = True
        else:
            self._mouse_viewport_move = False

    def handle_wheel_evt(self, delta: float) -> None:
        if delta > 0:
            self.meter_to_pixel += self.meter_to_pixel / 4
        else:
            self.meter_to_pixel -= self.meter_to_pixel / 4

    def show(self):
        while self._gui.running:

            for e in self._gui.get_events():
                if e.key == ti.GUI.ESCAPE:
                    exit()

                elif e.key == ti.GUI.SPACE and e.type == ti.GUI.RELEASE:
                    pass

                elif e.key == ti.GUI.LMB:
                    pass

                elif e.key == ti.GUI.RMB:
                    self.handle_right_mouse_evt(e.type)

                elif e.key == ti.GUI.MOVE:
                    self.handle_mouse_move_evt(e.pos[0], e.pos[1])

                elif e.key == ti.GUI.WHEEL:
                    self.handle_wheel_evt(e.delta[1])

                elif e.key == ti.GUI.UP:
                    print("press up key")

                elif e.key == ti.GUI.DOWN:
                    pass

                elif e.key == ti.GUI.LEFT and e.type == ti.GUI.RELEASE:
                    pass

                elif e.key == ti.GUI.RIGHT and e.type == ti.GUI.RELEASE:
                    pass

            self.physics_sim()
            self.render()
            self._gui.show()

    @property
    def meter_to_pixel(self) -> float:
        return self._meter_to_pixel

    @meter_to_pixel.setter
    def meter_to_pixel(self, val: float) -> None:
        # set the target 'meter_to_pixel' value
        # when in render,
        # clamp the scale value
        if val < 1.0:
            self._target_meter_to_pixel = 1.0
            self._target_pixel_to_meter = 1.0
            return

        self._target_meter_to_pixel = val
        self._target_pixel_to_meter = 1.0 / val

    def smooth_scale(self) -> None:
        # calc the 'meter to pixel' scale according
        # to the 'target meter to pixel' set from
        # the wheel event smooth animation
        inv_dt: float = 1.0 / self._delta_time
        scale: float = self._target_meter_to_pixel - self._meter_to_pixel
        if np.fabs(scale) < 0.1 or self._meter_to_pixel < 1.0:
            self._meter_to_pixel = self._target_meter_to_pixel
        else:
            self._meter_to_pixel -= (1.0 -
                                     ti.exp(self._restit * inv_dt)) * scale

        self._pixel_to_meter = 1.0 / self._meter_to_pixel



import os
import glob


class ExportManager():
    def __init__(self, root_dir: str = './export-res'):
        try:
            os.makedirs(root_dir + '/frames')
        except FileExistsError:
            for infile in glob.glob(os.path.join(root_dir + '/frames',
                                                 '*.png')):
                os.remove(infile)

        self._frame_cnt: int = -1
        self._root_dir: str = root_dir

    @property
    def frame_name(self) -> str:
        self._frame_cnt += 1
        return self._root_dir + f'/frames/{self._frame_cnt:05d}.png'

    def gen_video(self) -> None:
        print('export .mp4 video...')
        os.chdir(self._root_dir + '/frames')
        os.system('ti video -f24')
        os.system('mv video.mp4 ../')

    def gen_gif(self) -> None:
        print('export .gif ...')
        os.chdir(self._root_dir + '/frames')
        os.system('ti video -f24')
        os.system('ti gif -i video.mp4 -f24')
        os.system('mv video.gif ../')



from typing import List, Dict, Optional, Tuple

import numpy as np

from ...math.matrix import Matrix
from ...common.config import Config
from .gjk import PointPair
from ...geometry.geom_algo import GeomAlgo2D
from ...geometry.shape import Circle, Edge, Ellipse, Polygon, Shape, ShapePrimitive


class ProjectedPoint():
    def __init__(self):
        self._vertex: Matrix = Matrix([0.0, 0.0], 'vec')
        self._val: float = 0.0
        self._idx: int = -1

    def __eq__(self, other) -> bool:
        return self._vertex == other._vertex and np.isclose(
            self._val, other._val)


class ProjectedEdge():
    def __init__(self):
        self._vertex1 = Matrix([0.0, 0.0], 'vec')
        self._vertex2 = Matrix([0.0, 0.0], 'vec')


class ProjectedSegment():
    def __init__(self):
        self._min = ProjectedPoint()
        self._max = ProjectedPoint()

    @staticmethod
    def intersect(s1, s2):
        diff: float = Config.NegativeMin
        res = ProjectedSegment()

        if s1._min._val <= s2._min._val and s1._max._val <= s2._max._val:
            diff = s1._max._val - s2._min._val
            res._max = s1._max
            res._min = s2._min

        elif s2._min._val <= s1._min._val and s2._max._val <= s1._max._val:
            diff = s2._max._val - s1._min._val
            res._max = s2._max
            res._min = s1._min

        elif s1._min._val >= s2._min._val and s1._max._val <= s2._max._val:

            if (s1._max._val - s2._min._val) > (s2._max._val - s1._min._val):

                diff = s1._max._val - s2._min._val
                res._max = s1._max
                res._min = s2._min

            else:

                diff = s2._max._val - s1._min._val
                res._max = s2._max
                res._min = s1._min

        elif s2._min._val >= s1._min._val and s2._max._val <= s1._max._val:

            if (s2._max._val - s1._min._val) > (s1._max._val - s2._min._val):

                diff = s2._max._val - s1._min._val
                res._max = s2._max
                res._min = s1._min

            else:

                diff = s1._max._val - s2._min._val
                res._max = s1._max
                res._min = s2._min

        return (res, diff)


class SATResult():
    def __init__(self):
        self._contact_pair: List[PointPair] = []
        self._contact_pair_count: int = 0
        self._normal: Matrix = Matrix([0.0, 0.0], 'vec')
        self._penetration: float = 0.0
        self._is_colliding: bool = False


class SAT():
    @staticmethod
    def circle_vs_capsule(prima: ShapePrimitive,
                          primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Circle
        assert primb._shape.type() == Shape.Type.Capsule
        res: SATResult = SATResult()
        return res

    @staticmethod
    def circle_vs_sector(prima: ShapePrimitive,
                         primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Circle
        assert primb._shape.type() == Shape.Type.Sector
        res: SATResult = SATResult()
        return res

    @staticmethod
    def circle_vs_edge(prima: ShapePrimitive,
                       primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Circle
        assert primb._shape.type() == Shape.Type.Edge

        res: SATResult = SATResult()
        circle: Circle = prima._shape
        edg: Edge = primb._shape

        actual_start: Matrix = primb._xform + edg.start
        actual_end: Matrix = primb._xform + edg.end
        normal: Matrix = (actual_start - actual_end).normal()

        if (actual_start - prima._xform).dot(normal) < 0 and (
                actual_end - primb._xform).dot(normal) < 0:
            normal.negate()

        proj_point: Matrix = GeomAlgo2D.point_to_line_segment(
            actual_start, actual_end, prima._xform)
        diff: Matrix = proj_point - prima._xform

        res._normal = diff.normal()
        length: float = diff.len()

        res._is_colliding = (length < circle.radius)
        res._penetration = circle.radius - length
        res._contact_pair[0]._pa = prima._xform + circle.radius * res._normal
        res._contact_pair[0]._pb = proj_point
        res._contact_pair_count += 1

        return res

    @staticmethod
    def circle_vs_circle(prima: ShapePrimitive,
                         primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Circle
        assert primb._shape.type() == Shape.Type.Circle

        res: SATResult = SATResult()
        cira: Circle = prima._shape
        cirb: Circle = primb._shape

        ba: Matrix = prima._xform - primb._xform
        dp: float = cira.radius + cirb.radius
        length: float = ba.len()

        if length <= dp:
            res._normal = ba.normal()
            res._penetration = dp - length
            res._is_colliding = True
            res._contact_pair[0]._pa = prima._xform - cira.radius * res._normal
            res._contact_pair[0]._pb = primb._xform + cirb.radius * res._normal
            res._contact_pair_count += 1

        return res

    @staticmethod
    def circle_vs_polygon(prima: ShapePrimitive,
                          primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Circle
        assert primb._shape.type() == Shape.Type.Polygon

        cira: Circle = prima._shape
        polyb: Polygon = primb._shape
        colliding_axis: int = 0
        res: SATResult = SATResult()

        len_min: float = Config.Max
        closest: Matrix = Matrix([0.0, 0.0], 'vec')

        for elem in polyb.vertices:
            vertex: Matrix = primb.translate(elem)
            length: float = (vertex - prima._xform).len_square()
            if len_min > length:
                len_min = length
                closest = vertex

        normal: Matrix = closest.normal()
        seg_cir: ProjectedSegment = SAT._axis_projection(prima, cira, normal)
        seg_poly: ProjectedSegment = SAT._axis_projection(primb, polyb, normal)

        (final_seg, length) = ProjectedSegment.intersect(seg_cir, seg_poly)

        if length > 0:
            colliding_axis += 1

        if res._penetration > length:
            res._penetration = length
            res._normal = normal

        cirp: ProjectedPoint = final_seg._max if seg_cir._max == final_seg._max else final_seg._min

        vert_len: int = len(polyb.vertices)
        for i in range(vert_len - 1):
            v1: Matrix = primb.translate(polyb.vertices[i])
            v2: Matrix = primb.translate(polyb.vertices[i + 1])
            edg: Matrix = v1 - v2
            normal: Matrix = edg.perpendicular().normal()

            segc = SAT._axis_projection(prima, cira, normal)
            segp = SAT._axis_projection(primb, polyb, normal)

            (tmp_segment, tmp_length) = ProjectedSegment.intersect(segc, segp)
            if tmp_length > 0:
                colliding_axis += 1

            if res._penetration > tmp_length and tmp_length > 0:
                res._penetration = tmp_length
                res._normal = normal
                cirp = tmp_segment._max if segc._max == tmp_segment._max else tmp_segment._min

        if colliding_axis == len(polyb.vertices):
            res._is_colliding = True

        res._contact_pair[0]._pa = cirp._vertex
        res._contact_pair[
            0]._pb = cirp._vertex + -res._normal * res._penetration
        res._contact_pair_count += 1

        return res

    @staticmethod
    def polygon_vs_polygon(prima: ShapePrimitive,
                           primb: ShapePrimitive) -> SATResult:
        def _test(prima: ShapePrimitive, primb: ShapePrimitive) -> SATResult:
            polya: Polygon = prima._shape
            polyb: Polygon = primb._shape

            final_normal: Matrix = Matrix([0.0, 0.0], 'vec')
            len_min: float = Config.Max
            colliding_axis: int = 0

            tgtap: ProjectedPoint = ProjectedPoint()
            tgtbp: ProjectedPoint = ProjectedPoint()

            polya_len: int = len(polya.vertices)
            for i in range(polya_len - 1):
                v1: Matrix = prima.translate(polya.vertices[i])
                v2: Matrix = prima.translate(polya.vertices[i + 1])
                edg: Matrix = v1 - v2
                normal: Matrix = edg.perpendicular().normal()

                sega: ProjectedSegment = SAT._axis_projection(
                    prima, polya, normal)
                segb: ProjectedSegment = SAT._axis_projection(
                    primb, polyb, normal)

                (final_seg, length) = ProjectedSegment.intersect(sega, segb)
                if length > 0:
                    colliding_axis += 1

                polyap: ProjectedPoint = final_seg._max if sega._max == final_seg._max else final_seg._min
                polybp: ProjectedPoint = final_seg._max if segb._max == final_seg._max else final_seg._min

                if len_min > length:
                    len_min = length
                    final_normal = normal
                    tgtap = polyap
                    tgtbp = polybp

            return (final_normal, len_min, colliding_axis, tgtap, tgtbp)

        assert prima._shape.type() == Shape.Type.Polygon
        assert primb._shape.type() == Shape.Type.Polygon

        polya: Polygon = prima._shape
        polyb: Polygon = primb._shape

        res: SATResult = SATResult()
        (normal1, length1, axis1, polyap1, polybp1) = _test(prima, primb)
        (normal2, length2, axis2, polybp2, polyap2) = _test(primb, prima)

        if axis1 + axis2 == len(polya.vertices) + len(polyb.vertices) - 2:
            res._is_colliding = True

        if length1 < length2:
            res._penetration = length1
            res._normal = normal1
        else:
            res._penetration = length2
            res.normal = normal2

    @staticmethod
    def polygon_vs_edge(prima: ShapePrimitive,
                        primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Polygon
        assert primb._shape.type() == Shape.Type.Edge
        return SATResult()

    @staticmethod
    def polygon_vs_capsule(prima: ShapePrimitive,
                           primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Polygon
        assert primb._shape.type() == Shape.Type.Capsule
        return SATResult()

    @staticmethod
    def polygon_vs_sector(prima: ShapePrimitive,
                          primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Polygon
        assert primb._shape.type() == Shape.Type.Sector
        return SATResult()

    @staticmethod
    def capsule_vs_edge(prima: ShapePrimitive,
                        primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Capsule
        assert primb._shape.type() == Shape.Type.Edge
        return SATResult()

    @staticmethod
    def capsule_vs_capsule(prima: ShapePrimitive,
                           primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Capsule
        assert primb._shape.type() == Shape.Type.Capsule
        return SATResult()

    @staticmethod
    def capsule_vs_sector(prima: ShapePrimitive,
                          primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Capsule
        assert primb._shape.type() == Shape.Type.Sector
        return SATResult()

    @staticmethod
    def sector_vs_sector(prima: ShapePrimitive,
                         primb: ShapePrimitive) -> SATResult:
        assert prima._shape.type() == Shape.Type.Sector
        assert primb._shape.type() == Shape.Type.Sector
        return SATResult()

    @staticmethod
    def _axis_projection(prim: ShapePrimitive, shape: Shape,
                         normal: Matrix) -> ProjectedSegment:
        if shape.type() == Shape.Type.Polygon:
            point_min: ProjectedPoint = ProjectedPoint()
            point_max: ProjectedPoint = ProjectedPoint()

            point_min._val = Config.Max
            point_max._val = Config.NegativeMin

            shape_len: int = len(shape.vertices)
            for i in range(shape_len):
                vertex: Matrix = prim.translate(shape.vertices[i])
                value: float = vertex.dot(normal)

                if value < point_min._val:
                    point_min._vertex = vertex
                    point_min._val = value
                    point_min._idx = i

                if value > point_max._val:
                    point_max._vertex = vertex
                    point_max._val = value
                    point_max._idx = i

            segment: ProjectedSegment = ProjectedSegment()
            segment._max = point_max
            segment._min = point_min

            return segment

        elif shape.type() == Shape.Type.Circle:
            cir_min: ProjectedPoint = ProjectedPoint()
            cir_max: ProjectedPoint = ProjectedPoint()

            cir_min._vertex = prim._xform - normal * shape.radius
            cir_min._val = prim._xform.dot(normal) - shape.radius

            cir_max._vertex = prim._xform + normal * shape.radius
            cir_max._val = prim._xform.dot(normal) + shape.radius

            segment: ProjectedSegment = ProjectedSegment()
            segment._min = cir_min
            segment._max = cir_max

            return segment

        elif shape.type() == Shape.Type.Ellipse:
            elli_min: ProjectedPoint = ProjectedPoint()
            elli_max: ProjectedPoint = ProjectedPoint()

            rot_dir: Matrix = Matrix.rotate_mat(-prim._rot) * -normal
            elli_min._vertex = GeomAlgo2D.calc_ellipse_project_on_point(
                shape.A(), shape.B(), rot_dir)
            elli_min._vertex = prim.translate(elli_min._vertex)
            elli_min._val = elli_min._vertex.dot(normal)

            rot_dir = Matrix.rotate_mat(-prim._rot) * normal
            elli_max._vertex = GeomAlgo2D.calc_ellipse_project_on_point(
                shape.A(), shape.B(), rot_dir)
            elli_max._vertex = prim.translate(elli_max._vertex)
            elli_max._val = elli_max._vertex.dot(normal)

            segment: ProjectedSegment = ProjectedSegment()
            segment._min = elli_min
            segment._max = elli_max

            return segment

        elif shape.type() == Shape.Type.Capsule:
            capsule_min: ProjectedPoint = ProjectedPoint()
            capsule_max: ProjectedPoint = ProjectedPoint()

            dirn: Matrix = Matrix.rotate_mat(-prim._rot) * normal
            p1: Matrix = GeomAlgo2D.calc_capsule_project_on_point(
                shape.width, shape.height, dirn)
            p2: Matrix = GeomAlgo2D.calc_capsule_project_on_point(
                shape.width, shape.height, -dirn)
            p1 = prim.translate(p1)
            p2 = prim.translate(p2)

            capsule_min._vertex = p2
            capsule_min._val = capsule_min._vertex.dot(normal)

            capsule_max._vertex = p1
            capsule_max._val = capsule_max._vertex.dot(normal)

            segment: ProjectedSegment = ProjectedSegment()
            segment._min = capsule_min
            segment._max = capsule_max

            return segment

        elif shape.type() == Shape.Type.Sector:
            sector_min: ProjectedPoint = ProjectedPoint()
            sector_max: ProjectedPoint = ProjectedPoint()

            dirn: Matrix = Matrix.rotate_mat(-prim._rot) * normal
            p1: Matrix = GeomAlgo2D.calc_sector_project_on_point(
                shape.start, shape.span, shape.radius, dirn)
            p2: Matrix = GeomAlgo2D.calc_sector_project_on_point(
                shape.start, shape.span, shape.radius, -dirn)
            p1 = prim.translate(p1)
            p2 = prim.translate(p2)

            sector_min._vertex = p2
            sector_min._val = sector_min._vertex.dot(normal)

            sector_max._vertex = p1
            sector_max._val = sector_max._vertex.dot(normal)

            segment: ProjectedSegment = ProjectedSegment()
            segment._min = sector_min
            segment._max = sector_max

            return segment



from typing import List, Optional

import numpy as np

from ..math.matrix import Matrix
from ..dynamics.body import Body
from .algorithm.clip import ContactGenerator
from ..geometry.shape import ShapePrimitive
from ..collision.algorithm.gjk import PenetrationInfo, PenetrationSource
from ..collision.algorithm.gjk import GJK, PointPair


class Collsion():
    def __init__(self):
        self._is_colliding: bool = False
        self._bodya: Optional[Body] = None
        self._bodyb: Optional[Body] = None
        self._contact_list: List[PointPair] = []
        self._normal: Matrix = Matrix([0.0, 0.0], 'vec')
        self._penetration: float = 0.0


class Detector():
    @staticmethod
    def collide(bodya: Body, bodyb: Body) -> bool:
        assert bodya is not None
        assert bodyb is not None

        prima: ShapePrimitive = ShapePrimitive()
        primb: ShapePrimitive = ShapePrimitive()

        prima._shape = bodya.shape
        prima._rot = bodya.rot
        prima._xform = bodya.pos

        primb._shape = bodyb.shape
        primb._rot = bodyb.rot
        primb._xform = bodyb.pos

        (is_colliding, simplex) = GJK.gjk(prima, primb)
        if prima._xform == primb._xform and not is_colliding:
            is_colliding = simplex.contain_origin(True)

        return is_colliding

    @staticmethod
    def detect(bodya: Body, bodyb: Body) -> Collsion:
        res: Collsion = Collsion()

        if bodya is None or bodyb is None:
            return res

        if bodya == bodyb:
            return res

        if bodya.id > bodyb.id:
            bodya, bodyb = bodyb, bodya

        res._bodya = bodya
        res._bodyb = bodyb

        prima: ShapePrimitive = ShapePrimitive()
        primb: ShapePrimitive = ShapePrimitive()

        prima._shape = bodya.shape
        prima._rot = bodya.rot
        prima._xform = bodya.pos

        primb._shape = bodyb.shape
        primb._rot = bodyb.rot
        primb._xform = bodyb.pos

        (is_colliding, simplex) = GJK.gjk(prima, primb)
        if prima._xform == primb._xform and not is_colliding:
            is_colliding = simplex.contain_origin(True)

        res._is_colliding = is_colliding

        if is_colliding:
            # old_simplex: Simplex = simplex
            simplex = GJK.epa(prima, primb, simplex)
            source: PenetrationSource = GJK.dump_source(simplex)
            info: PenetrationInfo = GJK.dump_info(source)

            res._normal = info._normal
            res._penetration = info._penetration

            (clip_edga,
             clip_edgb) = ContactGenerator.recognize(prima, primb,
                                                     info._normal)

            pair_list: List[PointPair] = ContactGenerator.clip(
                clip_edga, clip_edgb, info._normal)

            val_pass: bool = False
            for elem in pair_list:
                tmp_val: float = (elem._pa - elem._pb).len_square()
                if np.isclose(tmp_val, res._penetration * res._penetration):
                    val_pass = True

            # if fail, there must be a deeper contact point, use it:
            if val_pass:
                res._contact_list = pair_list
            else:
                res._contact_list.append(GJK.dump_points(source))

        assert len(res._contact_list) != 3
        return res

    @staticmethod
    def distance(bodya: Body, bodyb: Body) -> PointPair:
        res: PointPair = PointPair()

        if bodya is None or bodyb is None:
            return res

        if bodya == bodyb:
            return res

        prima: ShapePrimitive = ShapePrimitive()
        primb: ShapePrimitive = ShapePrimitive()

        prima._shape = bodya.shape
        prima._rot = bodya.rot
        prima._xform = bodya.pos

        primb._shape = bodyb.shape
        primb._rot = bodyb.rot
        primb._xform = bodyb.pos
        return GJK.distance(prima, primb)



from __future__ import annotations
from typing import List, Optional, Tuple, cast

import numpy as np

from ...math.matrix import Matrix
from ...common.config import Config
from ...geometry.geom_algo import GeomAlgo2D
from ...geometry.shape import Capsule, Circle, Edge, Ellipse, Point
from ...geometry.shape import Polygon, Sector, Shape, ShapePrimitive


class Minkowski():
    def __init__(self,
                 pa: Matrix = Matrix([0.0, 0.0], 'vec'),
                 pb: Matrix = Matrix([0.0, 0.0], 'vec')):
        self._pa: Matrix = pa
        self._pb: Matrix = pb
        self._res: Matrix = self._pa - self._pb

    def __eq__(self, other) -> bool:
        return self._pa == other._pa and self._pb == other._pb

    def __ne__(self, other) -> bool:
        return not (self._pa == other._pa and self._pb == other._pb)


class Simplex():
    '''simplex structure for gjk/epa test.

    By convention:
    1 points:   p0, construct a single point
    2 points:   p0->pa, construct a segment
    >=4 points: p0->pa->pb->p0, construct a polygon
    '''
    def __init__(self):
        self._is_contain_origin: bool = False
        self._vertices: List[Minkowski] = []

    def contain_origin(self, strict: bool = False) -> bool:
        self._is_contain_origin = Simplex._contain_origin(self, strict)
        return self._is_contain_origin

    def insert(self, pos: int, vertex: Minkowski) -> None:
        self._vertices.insert(pos + 1, vertex)

    def contains(self, minkowski: Minkowski) -> bool:
        for v in self._vertices:
            if v == minkowski:
                return True

        return False

    def last_vertex(self) -> Matrix:
        vert_len: int = len(self._vertices)
        if vert_len == 2:
            return self._vertices[vert_len - 1]._res

        return self._vertices[vert_len - 2]._res

    @staticmethod
    def _contain_origin(simplex, strict: bool = False) -> bool:
        vert_len: int = len(simplex._vertices)
        if vert_len == 4:
            return GeomAlgo2D.is_triangle_contain_origin(
                simplex._vertices[0]._res, simplex._vertices[1]._res,
                simplex._vertices[2]._res)

        elif vert_len == 2:
            oa: Matrix = -simplex._vertices[0]._res
            ob: Matrix = -simplex._vertices[1]._res
            return GeomAlgo2D.is_point_on_segment(oa, ob,
                                                  Matrix([0.0, 0.0], 'vec'))

        else:
            return False


class PenetrationInfo():
    def __init__(self):
        self._normal: Matrix = Matrix([0.0, 0.0], 'vec')
        self._penetration: float = 0.0


class PenetrationSource():
    def __init__(self):
        self._a1: Matrix = Matrix([0.0, 0.0], 'vec')
        self._a2: Matrix = Matrix([0.0, 0.0], 'vec')
        self._b1: Matrix = Matrix([0.0, 0.0], 'vec')
        self._b2: Matrix = Matrix([0.0, 0.0], 'vec')


class PointPair():
    def __init__(self):
        self._pa: Matrix = Matrix([0.0, 0.0], 'vec')
        self._pb: Matrix = Matrix([0.0, 0.0], 'vec')

    def __eq__(self, other) -> bool:
        return self._pa == other._pa and self._pb == other._pb

    def __ne__(self, other) -> bool:
        return not (self._pa == other._pa and self._pb == other._pb)

    def is_empty(self) -> bool:
        return self._pa == Matrix([0.0, 0.0], 'vec') and self._pb == Matrix(
            [0.0, 0.0], 'vec')


class GJK():
    @staticmethod
    def gjk(prima: ShapePrimitive,
            primb: ShapePrimitive,
            iter_val: int = 20) -> Tuple[bool, Simplex]:
        '''Gilbert-Johnson-Keerthi distance algorithm

        Parameters
        ----------
        prima : ShapePrimitive
            primitive a
        primb : ShapePrimitive
            primitive b
        iter_val : int, optional
            iter num, by default 20

        Returns
        -------
        Tuple[bool, Simplex]
            return initial simplex and whether collision exist
        '''
        simplex: Simplex = Simplex()
        is_found: bool = False
        dirn: Matrix = primb._xform - prima._xform

        if dirn == Matrix([0.0, 0.0], 'vec'):
            dirn.set_value([1.0, 1.0])

        diff: Minkowski = GJK.support(prima, primb, dirn)
        simplex._vertices.append(diff)
        dirn.negate()

        removed: List[Minkowski] = []
        for i in range(iter_val):
            diff = GJK.support(prima, primb, dirn)
            simplex._vertices.append(diff)
            # print(f'i: {i}')

            # build a close polygon
            if len(simplex._vertices) == 3:
                simplex._vertices.append(simplex._vertices[0])
                # for v in simplex._vertices:
                # print(f'vertice res: {v._res}')

            if simplex.last_vertex().dot(dirn) <= 0:
                break

            if simplex.contain_origin(True):
                is_found = True
                break

            # if not contain origin
            # find edg closest to origin
            # reconstruct simplex
            # find the point that is not belong to the edg closest to origin
            # if found, there is no more minkowski difference, exit loop
            # if not, add the point to the list
            (idx1, idx2) = GJK.find_edge_closest_to_origin(simplex)
            dirn = GJK.calc_direction_by_edge(simplex._vertices[idx1]._res,
                                              simplex._vertices[idx2]._res,
                                              True)

            # print(f'idx1: {idx1} idx2: {idx2}')
            # print(f'dirn: {dirn}')
            res: Optional[Minkowski] = GJK.adjust_simplex(simplex, idx1, idx2)

            if res is not None:
                for v in removed:
                    if v == res:
                        break
                removed.append(res)

        return (is_found, simplex)

    @staticmethod
    def epa(prima: ShapePrimitive,
            primb: ShapePrimitive,
            src: Simplex,
            iter_val: int = 20) -> Simplex:
        '''Expanding Polygon Algorithm

        Parameters
        ----------
        prima : ShapePrimitive
            primitive a
        primb : ShapePrimitive
            primitive b
        src : Simplex
            init simplex
        iter_val : int, optional
            iter num, by default 20

        Returns
        -------
        Simplex
            return expanded simplex
        '''

        # edg: Simplex = Simplex()
        simplex: Simplex = src
        normal: Matrix = Matrix([0.0, 0.0], 'vec')
        p: Minkowski = Minkowski()

        for i in range(iter_val):
            (idx1, idx2) = GJK.find_edge_closest_to_origin(simplex)
            normal = GJK.calc_direction_by_edge(simplex._vertices[idx1]._res,
                                                simplex._vertices[idx2]._res,
                                                False).normal()

            if GeomAlgo2D.is_point_on_segment(simplex._vertices[idx1]._res,
                                              simplex._vertices[idx2]._res,
                                              Matrix([0.0, 0.0], 'vec')):
                normal.negate()

            p = GJK.support(prima, primb, normal)

            if simplex.contains(p):
                break

            simplex.insert(idx1, p)

        return simplex

    @staticmethod
    def dump_info(src: PenetrationSource) -> PenetrationInfo:
        '''Dump collision penetration normal and depth

        Parameters
        ----------
        src : PenetrationSource
            source data

        Returns
        -------
        PenetrationInfo
            collision info
        '''

        res: PenetrationInfo = PenetrationInfo()
        edg1: Matrix = src._a1 - src._b1
        edg2: Matrix = src._a2 - src._b2
        normal: Matrix = GJK.calc_direction_by_edge(edg1, edg2, False).normal()
        origin_to_edge: float = np.fabs(normal.dot(edg1))
        res._normal = normal.negate()
        res._penetration = origin_to_edge

        return res

    @staticmethod
    def support(prima: ShapePrimitive, primb: ShapePrimitive,
                dirn: Matrix) -> Minkowski:
        return Minkowski(GJK.find_farthest_point(prima, dirn),
                         GJK.find_farthest_point(primb, -dirn))

    @staticmethod
    def find_edge_closest_to_origin(simplex: Simplex) -> Tuple[int, int]:
        '''Find two points that can form an edge closest to origin of simplex

        Parameters
        ----------
        simplex : Simplex
            simplex

        Returns
        -------
        Tuple[int, int]
            the two idx of the point on the simplex match needed
        '''
        idx1: int = 0
        idx2: int = 0
        dist_min: float = Config.Max

        if len(simplex._vertices) == 2:
            return (0, 1)

        vert_len: int = len(simplex._vertices)
        for i in range(vert_len - 1):
            a: Matrix = simplex._vertices[i]._res
            b: Matrix = simplex._vertices[i + 1]._res
            p: Matrix = GeomAlgo2D.point_to_line_segment(
                a, b, Matrix([0.0, 0.0], 'vec'))

            proj: float = p.len()

            if dist_min > proj:
                idx1 = i
                idx2 = i + 1
                dist_min = proj

            elif np.isclose(dist_min, proj):
                length1: float = a.len_square() + b.len_square()
                length2: float = simplex._vertices[idx1]._res.len_square(
                ) + simplex._vertices[idx2]._res.len_square()

                if length1 < length2:
                    idx1 = i
                    idx2 = i + 1

        return (idx1, idx2)

    @staticmethod
    def find_farthest_point(prim: ShapePrimitive, dirn: Matrix) -> Matrix:
        '''Find farthest projection point in given direction

        Parameters
        ----------
        prim : ShapePrimitive
            primitive
        dirn : Matrix
            given direction

        Returns
        -------
        Matrix
            farthest point's val
        '''
        target: Matrix = Matrix([0.0, 0.0], 'vec')
        rot: Matrix = Matrix.rotate_mat(-prim._rot)
        rot_dir: Matrix = rot * dirn

        assert prim._shape is not None
        if prim._shape.type == Shape.Type.Polygon:
            poly: Polygon = cast(Polygon, prim._shape)
            (vertex, idx) = GJK.find_farthest_point2(poly.vertices, rot_dir)
            target = vertex

        elif prim._shape.type == Shape.Type.Circle:
            cir: Circle = cast(Circle, prim._shape)
            return dirn.normal() * cir.radius + prim._xform

        elif prim._shape.type == Shape.Type.Ellipse:
            elli: Ellipse = cast(Ellipse, prim._shape)
            target = GeomAlgo2D.calc_ellipse_project_on_point(
                elli.A(), elli.B(), rot_dir)

        elif prim._shape.type == Shape.Type.Edge:
            edg: Edge = cast(Edge, prim._shape)
            dot1: float = Matrix.dot_product(edg.start, dirn)
            dot2: float = Matrix.dot_product(edg.end, dirn)
            target = edg.start if dot1 > dot2 else edg.end

        elif prim._shape.type == Shape.Type.Point:
            point: Point = cast(Point, prim._shape)
            return point.pos

        elif prim._shape.type == Shape.Type.Capsule:
            cap: Capsule = cast(Capsule, prim._shape)
            target = GeomAlgo2D.calc_capsule_project_on_point(
                cap.width, cap.height, rot_dir)
            # print(f'capsule target: {target.x}, {target.y}')

        elif prim._shape.type == Shape.Type.Sector:
            sec: Sector = cast(Sector, prim._shape)
            target = GeomAlgo2D.calc_sector_project_on_point(
                sec.start, sec.span, sec.radius, rot_dir)

        # calc gemo algo in origin-base axis system
        # return the 'target' back in former coord
        rot.set_value(Matrix.rotate_mat(prim._rot))
        target = rot * target + prim._xform
        # print(f'target: {target}')
        return target

    @staticmethod
    def find_farthest_point2(vertices: List[Matrix],
                             dirn: Matrix) -> Tuple[Matrix, int]:
        '''find the farthest point of polygon
        in a given direction

        Parameters
        ----------
        vertices : List[Matrix]
            polygon's vertices
        dirn : Matrix
            given direction

        Returns
        -------
        Tuple[Matrix, int]
            pos and its index in polygon's vertices list
        '''
        val_max: float = Config.NegativeMin
        tgt_max: Matrix = Matrix([0.0, 0.0], 'vec')
        tgt_idx: int = 0
        vert_len: int = len(vertices)
        for i in range(vert_len):
            tmp: float = Matrix.dot_product(vertices[i], dirn)
            if val_max < tmp:
                val_max = tmp
                tgt_max = vertices[i]
                tgt_idx = i

        return (tgt_max, tgt_idx)

    @staticmethod
    def adjust_simplex(simplex: Simplex, closest1: int,
                       closest2: int) -> Optional[Minkowski]:
        '''Adjust triangle simplex, remove the point that can not form a
        triangle that contains origin

        Parameters
        ----------
        simplex : Simplex
            simplex
        closest1 : int
            closest idx1 on simplex
        closest2 : int
            closest idx2 on simplex

        Returns
        -------
        Optional[Minkowski]
            target minkowski
        '''

        vert_len: int = len(simplex._vertices)
        if vert_len == 4:
            idx: int = -1
            for i in range(vert_len - 1):
                if i != closest1 and i != closest2:
                    idx = i

            target: Minkowski = simplex._vertices[idx]
            del simplex._vertices[idx]

            # NOTE: need to calc again, because del oper
            vert_len = len(simplex._vertices)
            del simplex._vertices[vert_len - 1]

            return target

        return None

    @staticmethod
    def calc_direction_by_edge(pa: Matrix,
                               pb: Matrix,
                               point_to_origin: bool = True) -> Matrix:
        '''Given two points, calculate the perpendicular vector and
        the orientation is user-defined.

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        point_to_origin : bool, optional
            if point the origin, by default True

        Returns
        -------
        Matrix
            perpendicular vector
        '''

        ao: Matrix = -pa
        ab: Matrix = pb - pa
        perp_of_ab: Matrix = ab.perpendicular()

        if (Matrix.dot_product(ao, perp_of_ab) < 0
                and point_to_origin) or (Matrix.dot_product(ao, perp_of_ab) > 0
                                         and not point_to_origin):
            perp_of_ab.negate()

        return perp_of_ab

    @staticmethod
    def distance(prima: ShapePrimitive,
                 primb: ShapePrimitive,
                 iter_val: int = 20) -> PointPair:
        '''Calculate the distance of two shape primitive

        Parameters
        ----------
        prima : ShapePrimitive
            primitive a
        primb : ShapePrimitive
            primitive b
        iter_val : int, optional
            iter num, by default 20

        Returns
        -------
        PointPair
            point result
        '''

        simplex: Simplex = Simplex()
        dirn: Matrix = primb._xform - prima._xform

        # calc two minkowski
        m: Minkowski = GJK.support(prima, primb, dirn)
        simplex._vertices.append(m)
        dirn.negate()
        m = GJK.support(prima, primb, dirn)
        simplex._vertices.append(m)

        for i in range(iter_val):
            dirn = GJK.calc_direction_by_edge(simplex._vertices[0]._res,
                                              simplex._vertices[1]._res, True)
            m = GJK.support(prima, primb, dirn)

            if simplex.contains(m):
                break

            simplex._vertices.append(m)
            simplex._vertices.append(simplex._vertices[0])
            (idx1, idx2) = GJK.find_edge_closest_to_origin(simplex)
            GJK.adjust_simplex(simplex, idx1, idx2)

        return GJK.dump_points(GJK.dump_source(simplex))

    @staticmethod
    def dump_source(simplex: Simplex) -> PenetrationSource:
        res: PenetrationSource = PenetrationSource()

        (idx1, idx2) = GJK.find_edge_closest_to_origin(simplex)
        res._a1 = simplex._vertices[idx1]._pa
        res._a2 = simplex._vertices[idx2]._pa
        res._b1 = simplex._vertices[idx1]._pb
        res._b2 = simplex._vertices[idx2]._pb

        return res

    @staticmethod
    def dump_points(src: PenetrationSource) -> PointPair:
        res: PointPair = PointPair()

        a_s1: Matrix = src._a1
        b_s1: Matrix = src._a2
        a_s2: Matrix = src._b1
        b_s2: Matrix = src._b2

        a: Matrix = src._a1 - src._b1
        b: Matrix = src._a2 - src._b2
        lval: Matrix = b - a
        ll: float = lval.dot(lval)
        la: float = lval.dot(a)
        lambda2: float = -la / ll
        lambda1: float = 1 - lambda2

        res._pa.set_value(a_s1 * lambda1 + b_s1 * lambda2)
        res._pb.set_value(a_s2 * lambda1 + b_s2 * lambda2)

        if lval == Matrix([0.0, 0.0], 'vec') or lambda2 < 0:
            res._pa.set_value([a_s1.x, a_s1.y])
            res._pb.set_value([a_s2.x, a_s2.y])

        if lambda1 < 0:
            res._pa.set_value([b_s1.x, b_s1.y])
            res._pb.set_value([b_s2.x, b_s2.y])

        return res



import numpy as np

from ...math.matrix import Matrix
from ..body import Body
from .joint import Joint, JointType


class RotationJointPrimitive():
    def __init__(self):
        self._bodya: Body = Body()
        self._bodyb: Body = Body()
        self._ref_rot: float = 0.0
        self._eff_mass: float = 0.0
        self._bias: float = 0.0


class OrientationJointPrimitive():
    def __init__(self):
        self._bodya: Body = Body()
        self._target_point: Matrix = Matrix([0.0, 0.0], 'vec')
        self._ref_rot: float = 0.0
        self._eff_mass: float = 0.0
        self._bias: float = 0.0


class RotationJoint(Joint):
    def __init__(self,
                 prim: RotationJointPrimitive = RotationJointPrimitive()):
        super().__init__()
        self._type: JointType = JointType.Rotation
        self._prim: RotationJointPrimitive = prim
        self._factor: float = 0.2

    def set_value(self, prim: RotationJointPrimitive) -> None:
        self._prim = prim

    def prepare(self, dt: float) -> None:
        if self._prim._bodya is None or self._prim._bodyb is None:
            return

        ii_a: float = self._prim._bodya.inv_inertia
        ii_b: float = self._prim._bodyb.inv_inertia
        inv_dt: float = 1.0 / dt

        self._prim._eff_mass = 1.0 / (ii_a + ii_b)
        c: float = self._prim._bodya.rot - self._prim._bodyb.rot
        c -= self._prim._ref_rot  # NOTE: just for static check
        self._prim._bias = -self._factor * inv_dt * c

    def solve_velocity(self, dt: float) -> None:
        dw: float = self._prim._bodya.ang_vel - self._prim._bodyb.ang_vel
        impulse: float = self._prim._eff_mass * (-dw + self._prim._bias)

        self._prim._bodya.ang_vel += self._prim._bodya.inv_inertia * impulse
        self._prim._bodyb.ang_vel -= self._prim._bodyb.inv_inertia * impulse

    def solve_position(self, dt: float) -> None:
        pass

    def prim(self) -> RotationJointPrimitive:
        return self._prim


class OrientationJoint(Joint):
    def __init__(
        self, prim: OrientationJointPrimitive = OrientationJointPrimitive()):
        super().__init__()
        self._type: JointType = JointType.Orientation
        self._prim: OrientationJointPrimitive = prim
        self._factor: float = 1.0

    def set_value(self, prim: OrientationJointPrimitive) -> None:
        self._prim = prim

    def prepare(self, dt: float) -> None:
        if self._prim._bodya is None:
            return

        bodya: Body = self._prim._bodya
        point: Matrix = self._prim._target_point - bodya.pos
        target_rot: float = point.theta()

        ii_a: float = self._prim._bodya.inv_inertia
        inv_dt: float = 1.0 / dt
        self._prim._eff_mass = 1.0 / ii_a
        c: float = target_rot - self._prim._bodya.rot - self._prim._ref_rot

        if np.isclose(c, 2.0 * np.pi) or np.isclose(c, -2.0 * np.pi):
            c = 0
            bodya.rot = target_rot
            return

        self._prim._bias = self._factor * inv_dt * c

    def solve_velocity(self, dt: float) -> None:
        dw: float = self._prim._bodya.ang_vel
        print(f'rotation joint dw: {dw}')
        impulse: float = self._prim._eff_mass * (-dw * self._prim._bias)
        self._prim._bodya.ang_vel += self._prim._bodya.inv_inertia * impulse

    def solve_position(self, dt: float) -> None:
        pass

    def prim(self) -> OrientationJointPrimitive:
        return self._prim



from __future__ import annotations
from typing import Optional, Tuple, cast

import numpy as np

from ...common.config import Config
from ...math.matrix import Matrix
from ...geometry.geom_algo import GeomAlgo2D
from ...geometry.shape import Circle, Edge, Ellipse
from ...geometry.shape import Polygon, Shape, ShapePrimitive
from ...dynamics.body import Body
from ..algorithm.gjk import GJK


class AABB():
    def __init__(self, width: float = 0.0, height: float = 0.0):
        self._pos: Matrix = Matrix([0.0, 0.0], 'vec')
        self._width: float = width
        self._height: float = height

    def __eq__(self, other) -> bool:
        return np.isclose(self._width, other._width) and np.isclose(
            self._height, other._height) and self._pos == other._pos

    @property
    def pos(self) -> Matrix:
        return self._pos

    @pos.setter
    def pos(self, pos: Matrix) -> None:
        self._pos = pos

    @property
    def top_left(self) -> Matrix:
        return Matrix([
            -self._width / 2.0 + self._pos.x, self._height / 2.0 + self._pos.y
        ], 'vec')

    @property
    def top_right(self) -> Matrix:
        return Matrix([
            self._width / 2.0 + self._pos.x, self._height / 2.0 + self._pos.y
        ], 'vec')

    @property
    def bot_left(self) -> Matrix:
        return Matrix([
            -self._width / 2.0 + self._pos.x, -self._height / 2.0 + self._pos.y
        ], 'vec')

    @property
    def bot_right(self) -> Matrix:
        return Matrix([
            self._width / 2.0 + self._pos.x, -self._height / 2.0 + self._pos.y
        ], 'vec')

    def collide(self, aabb) -> bool:
        '''check if two aabbs are overlapping

        Parameters
        ----------
        aabb : AABB
            other AABB

        Returns
        -------
        bool
            True: collide, otherwise not
        '''
        return AABB._collide(self, aabb)

    def expand(self, factor: float) -> None:
        AABB._expand(self, factor)

    def scale(self, factor: float) -> None:
        self._width *= factor
        self._height *= factor

    def clear(self) -> None:
        self._pos.clear()
        self._width = 0.0
        self._height = 0.0

    def unite(self, aabb: AABB) -> AABB:
        '''return two AABB union result

        Parameters
        ----------
        aabb : AABB
            other AABB

        Returns
        -------
        AABB
            union AABB
        '''
        return AABB._unite(self, aabb)

    def surface_area(self) -> float:
        return (self._width + self._height) * 2.0

    def volume(self) -> float:
        return self._width * self._height

    def is_subset(self, aabb) -> bool:
        '''check if self is the subset of aabb

        Parameters
        ----------
        aabb : AABB
            ref AABB

        Returns
        -------
        bool
            True: is subset, otherwise not
        '''
        return AABB._is_subset(aabb, self)  # NOTE: the seq is important

    def is_empty(self) -> bool:
        return np.isclose(self._width, 0.0) and np.isclose(
            self._height, 0.0) and self._pos == Matrix([0.0, 0.0], 'vec')

    def raycast(self, start: Matrix, dirn: Matrix) -> bool:
        return AABB._raycast(self, start, dirn)

    @staticmethod
    def from_prim(prim: ShapePrimitive, factor: float = 0.0) -> AABB:
        res: AABB = AABB()
        assert prim._shape is not None

        if prim._shape.type == Shape.Type.Polygon:
            polygon: Polygon = cast(Polygon, prim._shape)
            x_max: float = Config.NegativeMin
            y_max: float = Config.NegativeMin
            x_min: float = Config.Max
            y_min: float = Config.Max
            for v in polygon.vertices:
                vertex: Matrix = Matrix.rotate_mat(prim._rot) * v

                if x_max < vertex.x:
                    x_max = vertex.x

                if x_min > vertex.x:
                    x_min = vertex.x

                if y_max < vertex.y:
                    y_max = vertex.y

                if y_min > vertex.y:
                    y_min = vertex.y

            res._width = np.fabs(x_max - x_min)
            res._height = np.fabs(y_max - y_min)
            res._pos.set_value([(x_max + x_min) / 2.0, (y_max + y_min) / 2.0])

        elif prim._shape.type == Shape.Type.Ellipse:
            ellipse: Ellipse = cast(Ellipse, prim._shape)

            top_dir: Matrix = Matrix([0.0, 1.0], 'vec')
            left_dir: Matrix = Matrix([-1.0, 0.0], 'vec')
            bot_dir: Matrix = Matrix([0.0, -1.0], 'vec')
            right_dir: Matrix = Matrix([1.0, 0.0], 'vec')

            top_dir = Matrix.rotate_mat(-prim._rot) * top_dir
            left_dir = Matrix.rotate_mat(-prim._rot) * left_dir
            bot_dir = Matrix.rotate_mat(-prim._rot) * bot_dir
            right_dir = Matrix.rotate_mat(-prim._rot) * right_dir

            top: Matrix = GeomAlgo2D.calc_ellipse_project_on_point(
                ellipse.A(), ellipse.B(), top_dir)
            left: Matrix = GeomAlgo2D.calc_ellipse_project_on_point(
                ellipse.A(), ellipse.B(), left_dir)
            bot: Matrix = GeomAlgo2D.calc_ellipse_project_on_point(
                ellipse.A(), ellipse.B(), bot_dir)
            right: Matrix = GeomAlgo2D.calc_ellipse_project_on_point(
                ellipse.A(), ellipse.B(), right_dir)

            top = Matrix.rotate_mat(prim._rot) * top
            left = Matrix.rotate_mat(prim._rot) * left
            bot = Matrix.rotate_mat(prim._rot) * bot
            right = Matrix.rotate_mat(prim._rot) * right

            res._height = np.fabs(top.y - bot.y)
            res._width = np.fabs(right.x - left.x)

        elif prim._shape.type == Shape.Type.Circle:
            cir: Circle = cast(Circle, prim._shape)
            res._width = cir.radius * 2.0
            res._height = cir.radius * 2.0

        elif prim._shape.type == Shape.Type.Edge:
            edg: Edge = cast(Edge, prim._shape)
            res._width = np.fabs(edg.start.x - edg.end.x)
            res._height = np.fabs(edg.start.y - edg.end.y)
            res._pos.set_value(
                [edg.start.x + edg.end.x, edg.start.y + edg.end.y])
            res._pos *= 0.5

        elif prim._shape.type == Shape.Type.Curve:
            pass
        elif prim._shape.type == Shape.Type.Point:
            res._width = 1.0
            res._height = 1.0

        elif prim._shape.type == Shape.Type.Capsule:
            p1: Matrix = GJK.find_farthest_point(prim, Matrix([1.0, 0.0],
                                                              'vec'))
            p2: Matrix = GJK.find_farthest_point(prim, Matrix([0.0, 1.0],
                                                              'vec'))

            p1 -= prim._xform
            p2 -= prim._xform
            res._width = p1.x * 2.0
            res._height = p2.y * 2.0
        elif prim._shape.type == Shape.Type.Sector:
            pass

        res._pos += prim._xform
        res.expand(factor)
        return res

    @staticmethod
    def from_body(body: Body, factor: float = 0.0) -> AABB:
        assert body is not None
        assert body.shape is not None

        prim: ShapePrimitive = ShapePrimitive()
        prim._shape = body.shape
        prim._rot = body.rot
        prim._xform = body.pos

        return AABB.from_prim(prim, factor)

    @staticmethod
    def from_box(top_left: Matrix, bot_right: Matrix) -> AABB:
        res: AABB = AABB()
        res._width = bot_right.x - top_left.x
        res._height = top_left.y - bot_right.y
        res._pos = (top_left + bot_right) / 2.0
        return res

    @staticmethod
    def _collide(src: AABB, target: AABB) -> bool:
        '''check if two aabbs are overlapping

        Parameters
        ----------
        src : AABB
            src AABB
        target : AABB
            target AABB

        Returns
        -------
        bool
            True: collide, otherwise not
        '''
        src_top_left: Matrix = src.top_left
        src_bot_right: Matrix = src.bot_right
        tgt_top_left: Matrix = target.top_left
        tgt_bot_right: Matrix = target.bot_right

        return not (src_bot_right.x < tgt_top_left.x
                    or tgt_bot_right.x < src_top_left.x
                    or src_top_left.y < tgt_bot_right.y
                    or tgt_top_left.y < src_bot_right.y)

    @staticmethod
    def _unite(src: AABB, target: AABB, factor: float = 0.0) -> AABB:
        '''return two AABB union result

        Parameters
        ----------
        src : AABB
            source AABB
        target : AABB
            target AABB
        factor : float
            expand factor

        Returns
        -------
        AABB
            union AABB
        '''

        if src.is_empty():
            return target
        elif target.is_empty():
            return src

        src_top_left: Matrix = src.top_left
        src_bot_right: Matrix = src.bot_right
        tgt_top_left: Matrix = target.top_left
        tgt_bot_right: Matrix = target.bot_right

        x_min: float = np.fmin(src_top_left.x, tgt_top_left.x)
        x_max: float = np.fmax(src_bot_right.x, tgt_bot_right.x)
        y_min: float = np.fmin(src_bot_right.y, tgt_bot_right.y)
        y_max: float = np.fmax(src_top_left.y, tgt_top_left.y)

        res: AABB = AABB()
        res._pos.set_value([(x_min + x_max) / 2.0, (y_min + y_max) / 2.0])
        res._width = x_max - x_min
        res._height = y_max - y_min
        res.expand(factor)

        return res

    @staticmethod
    def _is_subset(src: AABB, target: AABB) -> bool:
        '''check if target is the subset of src

        Parameters
        ----------
        src : AABB
            ref AABB
        target : AABB
            dut AABB

        Returns
        -------
        bool
            True: is subset, otherwise not
        '''
        src_top_left: Matrix = src.top_left
        src_bot_right: Matrix = src.bot_right
        tgt_top_left: Matrix = target.top_left
        tgt_bot_right: Matrix = target.bot_right

        x_ck1: bool = src_bot_right.x >= tgt_bot_right.x
        x_ck2: bool = tgt_top_left.x >= src_top_left.x
        y_ck1: bool = src_top_left.y >= tgt_top_left.y
        y_ck2: bool = tgt_bot_right.y >= src_bot_right.y
        return x_ck1 and x_ck2 and y_ck1 and y_ck2

    @staticmethod
    def _expand(aabb: AABB, factor: float = 0.0) -> None:
        aabb._width += factor
        aabb._height += factor

    @staticmethod
    def _raycast(aabb: AABB, start: Matrix, dirn: Matrix) -> bool:
        res: Optional[Tuple[Matrix, Matrix]] = GeomAlgo2D.raycast_aabb(
            start, dirn, aabb.top_left, aabb.bot_right)

        if res is None:
            return False

        p1, p2 = res[0], res[1]
        return GeomAlgo2D.is_point_on_AABB(
            p1, aabb.top_left, aabb.bot_right) and GeomAlgo2D.is_point_on_AABB(
                p2, aabb.top_left, aabb.bot_right)



import sys
import numpy as np
import taichi as ti

# add the TaichiGAME lib to the path
# below code is only needed in dev condition
sys.path.append('../')

import TaichiGAME as ng

from TaichiGAME.ti_scene import Scene

ti.init(arch=ti.gpu, excepthook=True)

scene = Scene('GPU Testbed')


class FrameBroadPhaseDetect(ng.Frame):
    def load(self) -> None:
        tri_data = [
            ng.Matrix([-1.0, 1.0], 'vec'),
            ng.Matrix([0.0, -2.0], 'vec'),
            ng.Matrix([1.0, -1.0], 'vec'),
            ng.Matrix([-1.0, 1.0], 'vec'),
        ]

        poly_data = [
            ng.Matrix([0.0, 4.0], 'vec'),
            ng.Matrix([-3.0, 3.0], 'vec'),
            ng.Matrix([-4.0, 0.0], 'vec'),
            ng.Matrix([-3.0, -3.0], 'vec'),
            ng.Matrix([0.0, -4.0], 'vec'),
            ng.Matrix([0.0, 4.0], 'vec')
        ]

        rect: ng.Rectangle = ng.Rectangle(0.5, 0.5)
        cir: ng.Circle = ng.Circle(0.5)
        cap: ng.Capsule = ng.Capsule(1.5, 0.5)
        tri: ng.Polygon = ng.Polygon()
        tri.vertices = tri_data
        poly: ng.Polygon = ng.Polygon()
        poly.vertices = poly_data

        tri.scale(0.5)
        poly.scale(0.1)

        rng = np.random.default_rng(seed=6)

        bd = ng.Body()
        tmpx: float = 0.0
        tmpy: float = 0.0
        cnt = np.array([])
        for i in range(51):
            bd = scene._world.create_body()
            tmpx = -10 + rng.random() * 20.0
            tmpy = -6 + rng.random() * 12.0
            bd.pos = ng.Matrix([tmpx, tmpy], 'vec')

            cnt = rng.integers(low=0, high=5, size=1)
            if cnt[0] == 0:
                bd.shape = rect
            elif cnt[0] == 1:
                bd.shape = cir
            elif cnt[0] == 2:
                bd.shape = tri
            elif cnt[0] == 3:
                bd.shape = poly
            elif cnt[0] == 4:
                bd.shape = cap

            bd.rot = -np.pi + rng.random() * np.pi
            bd.mass = 1
            bd.type = ng.Body.Type.Static
            # scene._dbvt.insert(bd)

    def render(self) -> None:
        pass
        # render the query rectangle range
        # ng.Render.rd_rect(
        #     scene._gui,
        #     scene._cam.world_to_screen(ng.Matrix([-4.0, 4.0], 'vec')),
        #     scene._cam.world_to_screen(ng.Matrix([4.0, -4.0], 'vec')),
        #     color=ng.Config.QueryRectLineColor)

        # query_region = ng.AABB(8, 8)
        # query_body_list = scene._dbvt.query(query_region)

        # for bd in query_body_list:
        #     scene._cam.render_aabb(scene._gui, ng.AABB.from_body(bd))


frame_broad_phase = FrameBroadPhaseDetect()
scene.register_frame(frame_broad_phase)
scene.init_frame()
scene.show()



from .joint import Joint, JointType


class PulleyJointPrimitive():
    def __init__(self):
        pass


class PulleyJoint(Joint):
    def __init__(self, prim: PulleyJointPrimitive = PulleyJointPrimitive()):
        super().__init__()
        self._type: JointType = JointType.Pulley
        self._prim: PulleyJointPrimitive = prim

    def set_value(self, prim: PulleyJointPrimitive):
        self._prim = prim

    def prepare(self, dt: float) -> None:
        raise NotImplementedError

    def solve_velocity(self, dt: float) -> None:
        raise NotImplementedError

    def solve_position(self, dt: float) -> None:
        raise NotImplementedError



from typing import Callable, List, Optional, Tuple, cast

import numpy as np
import taichi as ti

from ..common.config import Config
from ..math.matrix import Matrix
from ..geometry.shape import Capsule, Circle, Edge
from ..geometry.shape import Polygon, Shape, ShapePrimitive
from ..dynamics.joint.joint import Joint, JointType
from ..dynamics.joint.distance import DistanceJoint
from ..dynamics.joint.point import PointJoint
from ..dynamics.joint.revolute import RevoluteJoint
from ..dynamics.joint.revolute import RevoluteJointPrimitive


class Render():
    @staticmethod
    def rd_point(gui: ti.GUI,
                 point: Matrix,
                 color: int = ti.rgb_to_hex([1.0, 1.0, 1.0]),
                 radius: float = 2.0) -> None:
        assert gui is not None
        # assert 0 <= point.x <= 1.0
        # assert 0 <= point.y <= 1.0
        gui.circle([point.x, point.y], color, radius)

    @staticmethod
    def rd_points(gui: ti.GUI,
                  points: List[Matrix],
                  color: int = ti.rgb_to_hex([1.0, 1.0, 1.0]),
                  radius: float = 2.0) -> None:
        assert gui is not None

        for p in points:
            Render.rd_point(gui, p, color, radius)

    @staticmethod
    def rd_line(gui: ti.GUI,
                p1: Matrix,
                p2: Matrix,
                color: int = ti.rgb_to_hex([1.0, 1.0, 1.0]),
                radius: float = 1.0) -> None:
        assert gui is not None
        # assert 0 <= p1.x <= 1.0
        # assert 0 <= p1.y <= 1.0
        # assert 0 <= p2.x <= 1.0
        # assert 0 <= p2.y <= 1.0

        gui.line([p1.x, p1.y], [p2.x, p2.y], radius, color)

    @staticmethod
    def rd_lines(gui: ti.GUI,
                 lines: List[Tuple[Matrix, Matrix]],
                 color: int = ti.rgb_to_hex([1.0, 1.0, 1.0]),
                 radius: float = 1.0) -> None:
        assert gui is not None

        for lin in lines:
            Render.rd_line(gui, lin[0], lin[1], color, radius)

    @staticmethod
    def rd_shape(gui: ti.GUI,
                 prim: ShapePrimitive,
                 world_to_screen: Callable[[Matrix], Matrix],
                 meter_to_pixel: float,
                 fill_color: int = Config.FillColor,
                 outline_color: int = Config.OuterLineColor) -> None:
        assert gui is not None
        assert prim._shape is not None

        if prim._shape.type == Shape.Type.Polygon:
            Render.rd_polygon(gui, prim, world_to_screen, fill_color,
                              outline_color)

        elif prim._shape.type == Shape.Type.Ellipse:
            Render.rd_ellipse()

        elif prim._shape.type == Shape.Type.Circle:
            Render.rd_circle(gui, prim, world_to_screen, meter_to_pixel,
                             fill_color)

        elif prim._shape.type == Shape.Type.Curve:
            Render.rd_curve()

        elif prim._shape.type == Shape.Type.Edge:
            Render.rd_edge(gui, prim, world_to_screen)

        elif prim._shape.type == Shape.Type.Capsule:
            Render.rd_capsule(gui, prim, world_to_screen, meter_to_pixel,
                              fill_color)

        elif prim._shape.type == Shape.Type.Sector:
            raise NotImplementedError

    @staticmethod
    def rd_polygon(gui: ti.GUI,
                   prim: ShapePrimitive,
                   world_to_screen: Callable[[Matrix], Matrix],
                   fill_color: int = Config.FillColor,
                   outline_color: int = Config.OuterLineColor) -> None:

        # [trick] draw polygon by draw multi triangle
        poly: Polygon = cast(Polygon, prim._shape)
        assert len(poly.vertices) >= 3

        outer_line_st: Optional[np.ndarray] = None
        outer_line_ed: Optional[np.ndarray] = None
        fill_tri_pa: Optional[np.ndarray] = None
        fill_tri_pb: Optional[np.ndarray] = None
        fill_tri_pc: Optional[np.ndarray] = None

        is_first: bool = True

        # NOTE: the vertices can form a close shape,
        # so the first vertex and last vertex are same
        # print('render poly')
        vert_len: int = len(poly.vertices)
        for i in range(vert_len - 1):
            wordpa: Matrix = Matrix.rotate_mat(
                prim._rot) * poly.vertices[i] + prim._xform
            scrnpa: Matrix = world_to_screen(wordpa)

            wordpb: Matrix = Matrix.rotate_mat(
                prim._rot) * poly.vertices[i + 1] + prim._xform
            scrnpb: Matrix = world_to_screen(wordpb)

            if is_first:
                is_first = False
                outer_line_st = np.array([[scrnpa.x, scrnpa.y]])
                outer_line_ed = np.array([[scrnpb.x, scrnpb.y]])
            else:
                tmpa = np.array([[scrnpa.x, scrnpa.y]])
                tmpb = np.array([[scrnpb.x, scrnpb.y]])
                outer_line_st = np.concatenate((outer_line_st, tmpa))
                outer_line_ed = np.concatenate((outer_line_ed, tmpb))

        assert outer_line_st is not None
        assert outer_line_ed is not None

        fill_tri_pa = np.repeat(np.array([outer_line_st[0]]),
                                vert_len - 3,
                                axis=0)
        fill_tri_pb = outer_line_st[1:-1]
        fill_tri_pc = outer_line_st[2:]

        # print('poly ver:')
        # for v in poly.vertices:
        # print(v)
        # print('end')
        # print(f'line_len: {len(outer_line_st)}')
        # for v in outer_line_st:
        # print(v)

        gui.lines(outer_line_st, outer_line_ed, 1.5, outline_color)
        gui.triangles(fill_tri_pa, fill_tri_pb, fill_tri_pc, fill_color)

    @staticmethod
    def rd_ellipse() -> None:
        raise NotImplementedError

    @staticmethod
    def rd_circle(gui: ti.GUI,
                  prim: ShapePrimitive,
                  world_to_screen: Callable[[Matrix], Matrix],
                  meter_to_pixel: float,
                  color: int = Config.FillColor) -> None:
        cir: Circle = cast(Circle, prim._shape)
        scrnp: Matrix = world_to_screen(prim._xform)
        gui.circle([scrnp.x, scrnp.y],
                   color,
                   radius=cir.radius * meter_to_pixel)

    @staticmethod
    def rd_curve() -> None:
        raise NotImplementedError

    @staticmethod
    def rd_edge(gui: ti.GUI, prim: ShapePrimitive,
                world_to_screen: Callable[[Matrix], Matrix]) -> None:
        edg: Edge = cast(Edge, prim._shape)
        tmp1: Matrix = world_to_screen(edg.start + prim._xform)
        tmp2: Matrix = world_to_screen(edg.end + prim._xform)
        Render.rd_point(gui, tmp1, Config.AxisPointColor)
        Render.rd_point(gui, tmp2, Config.AxisPointColor)
        Render.rd_line(gui, tmp1, tmp2, Config.AxisLineColor)

        center: Matrix = edg.center()
        center += prim._xform
        Render.rd_line(gui, world_to_screen(center),
                       world_to_screen(center + edg.normal * 0.1),
                       Config.AxisLineColor)

    @staticmethod
    def rd_capsule(gui: ti.GUI,
                   prim: ShapePrimitive,
                   world_to_screen: Callable[[Matrix], Matrix],
                   meter_to_pixel: float,
                   color: int = Config.FillColor) -> None:
        cap: Capsule = cast(Capsule, prim._shape)

        # render two tangent circle
        offset: float = np.fmin(cap.width, cap.height) / 2.0
        tmp1: Matrix = Matrix([0.0, 0.0], 'vec')
        tmp2: Matrix = Matrix([0.0, 0.0], 'vec')
        if cap.width > cap.height:
            tmp1.x = -(cap.width / 2.0 - offset)
            tmp2.x = cap.width / 2.0 - offset
        else:
            tmp1.y = -(cap.height / 2.0 - offset)
            tmp2.y = cap.height / 2.0 - offset

        tmp1 = Matrix.rotate_mat(prim._rot) * tmp1 + prim._xform
        tmp2 = Matrix.rotate_mat(prim._rot) * tmp2 + prim._xform
        tmp1 = world_to_screen(tmp1)
        tmp2 = world_to_screen(tmp2)

        gui.circle([tmp1.x, tmp1.y], color, radius=offset * meter_to_pixel)
        gui.circle([tmp2.x, tmp2.y], color, radius=offset * meter_to_pixel)

        # render inner rectangle
        capw: float = cap.width
        caph: float = cap.height
        rectp1: Matrix = Matrix([-capw / 2.0, -caph / 2.0], 'vec')
        rectp2: Matrix = Matrix([capw / 2.0, -caph / 2.0], 'vec')
        rectp3: Matrix = Matrix([capw / 2.0, caph / 2.0], 'vec')
        rectp4: Matrix = Matrix([-capw / 2.0, caph / 2.0], 'vec')

        if capw > caph:
            rectp1.x += offset
            rectp2.x -= offset
            rectp3.x -= offset
            rectp4.x += offset
        else:
            rectp1.y += offset
            rectp2.y += offset
            rectp3.y -= offset
            rectp4.y -= offset

        rectp1 = Matrix.rotate_mat(prim._rot) * rectp1 + prim._xform
        rectp2 = Matrix.rotate_mat(prim._rot) * rectp2 + prim._xform
        rectp3 = Matrix.rotate_mat(prim._rot) * rectp3 + prim._xform
        rectp4 = Matrix.rotate_mat(prim._rot) * rectp4 + prim._xform

        rectp1 = world_to_screen(rectp1)
        rectp2 = world_to_screen(rectp2)
        rectp3 = world_to_screen(rectp3)
        rectp4 = world_to_screen(rectp4)

        # render the center rectangle
        fill_tri_pa: np.ndarray = np.array([[rectp1.x, rectp1.y],
                                            [rectp1.x, rectp1.y]])
        fill_tri_pb: np.ndarray = np.array([[rectp2.x, rectp2.y],
                                            [rectp3.x, rectp3.y]])
        fill_tri_pc: np.ndarray = np.array([[rectp3.x, rectp3.y],
                                            [rectp4.x, rectp4.y]])

        gui.triangles(fill_tri_pa, fill_tri_pb, fill_tri_pc, color)

    @staticmethod
    def rd_rect(gui: ti.GUI,
                top_left: Matrix,
                bot_right: Matrix,
                color: int = ti.rgb_to_hex([1.0, 1.0, 1.0]),
                radius: float = 1.0) -> None:
        assert gui is not None
        # assert 0 <= top_left.x <= 1.0
        # assert 0 <= top_left.y <= 1.0
        # assert 0 <= bot_right.x <= 1.0
        # assert 0 <= bot_right.y <= 1.0

        gui.rect([top_left.x, top_left.y], [bot_right.x, bot_right.y], radius,
                 color)

    @staticmethod
    def rd_joint(gui: ti.GUI, joint: Joint,
                 world_to_screen: Callable[[Matrix], Matrix]) -> None:
        jt_type: JointType = joint.type()
        if jt_type == JointType.Rotation:
            Render.rd_rotation_joint()

        elif jt_type == JointType.Distance:
            Render.rd_distance_joint(gui, joint, world_to_screen)

        elif jt_type == JointType.Point:
            Render.rd_point_joint(gui, joint, world_to_screen)

        elif jt_type == JointType.Orientation:
            Render.rd_orient_joint(gui, joint, world_to_screen)

        elif jt_type == JointType.Pulley:
            Render.rd_pulley_joint()

        elif jt_type == JointType.Prismatic:
            Render.rd_prismatic_joint()

        elif jt_type == JointType.Revolute:
            Render.rd_revolute_joint(gui, joint, world_to_screen)

        elif jt_type == JointType.Wheel:
            Render.rd_wheel_joint()

    @staticmethod
    def rd_rotation_joint() -> None:
        pass

    @staticmethod
    def rd_distance_joint(gui: ti.GUI, joint: Joint,
                          world_to_screen: Callable[[Matrix], Matrix]) -> None:
        assert joint is not None

        dis_jt: DistanceJoint = cast(DistanceJoint, joint)

        assert dis_jt.prim._bodya is not None
        pa: Matrix = dis_jt.prim._bodya.to_world_point(
            dis_jt.prim._local_pointa)
        pb: Matrix = dis_jt.prim._target_point

        n: Matrix = (pa - pb).normal()
        point_min: Matrix = n * dis_jt.prim._dist_min + pb
        point_max: Matrix = n * dis_jt.prim._dist_max + pb

        Render.rd_point(gui, world_to_screen(pa), Config.JointPointColor)
        Render.rd_point(gui, world_to_screen(pb), Config.JointPointColor)
        Render.rd_point(gui, world_to_screen(point_min), 0x448AFF)
        Render.rd_point(gui, world_to_screen(point_max), 0xF44336)
        Render.rd_line(gui, world_to_screen(pa), world_to_screen(pb),
                       Config.JointLineColor)

    @staticmethod
    def rd_point_joint(gui: ti.GUI, joint: Joint,
                       world_to_screen: Callable[[Matrix], Matrix]) -> None:
        assert joint is not None

        point_jt: PointJoint = cast(PointJoint, joint)
        pa: Matrix = point_jt.prim()._bodya.to_world_point(
            point_jt.prim()._local_pointa)
        pb: Matrix = point_jt.prim()._target_point

        Render.rd_point(gui, world_to_screen(pa), Config.JointPointColor)
        Render.rd_point(gui, world_to_screen(pb), Config.JointPointColor)
        Render.rd_line(gui, world_to_screen(pa), world_to_screen(pb),
                       Config.JointLineColor)

    @staticmethod
    def rd_orient_joint(gui: ti.GUI, joint: Joint,
                        world_to_screen: Callable[[Matrix], Matrix]) -> None:
        assert joint is not None
        point_jt: PointJoint = cast(PointJoint, joint)
        pa: Matrix = point_jt.prim()._bodya.pos
        pb: Matrix = point_jt.prim()._target_point

        Render.rd_point(gui, world_to_screen(pa), Config.JointPointColor)
        Render.rd_point(gui, world_to_screen(pb), Config.JointPointColor)
        Render.rd_line(gui, world_to_screen(pa), world_to_screen(pb),
                       Config.JointLineColor)

    @staticmethod
    def rd_pulley_joint() -> None:
        pass

    @staticmethod
    def rd_prismatic_joint() -> None:
        pass

    @staticmethod
    def rd_revolute_joint(gui: ti.GUI, joint: Joint,
                          world_to_screen: Callable[[Matrix], Matrix]) -> None:
        assert joint is not None
        revol: RevoluteJoint = cast(RevoluteJoint, joint)

        tmp: RevoluteJointPrimitive = revol.prim()
        assert tmp._bodya is not None
        assert tmp._bodyb is not None
        pa: Matrix = tmp._bodya.to_world_point(tmp._local_pointa)
        pb: Matrix = tmp._bodyb.to_world_point(tmp._local_pointb)

        Render.rd_point(gui, world_to_screen(pa), Config.JointPointColor)
        Render.rd_point(gui, world_to_screen(pb), Config.JointPointColor)
        Render.rd_line(gui, world_to_screen(pa), world_to_screen(pb),
                       Config.JointLineColor)

    @staticmethod
    def rd_wheel_joint() -> None:
        pass

    @staticmethod
    def rd_angle_line(gui: ti.GUI, prim: ShapePrimitive,
                      world_to_screen: Callable[[Matrix], Matrix]) -> None:
        xpos: Matrix = Matrix([0.15, 0.0], 'vec')
        ypos: Matrix = Matrix([0.0, 0.15], 'vec')

        assert prim._shape is not None
        mc: Matrix = Matrix.rotate_mat(prim._rot) * prim._shape.center()
        start: Matrix = prim._xform + mc
        xpos = Matrix.rotate_mat(prim._rot) * xpos + start
        ypos = Matrix.rotate_mat(prim._rot) * ypos + start

        Render.rd_line(gui, world_to_screen(start), world_to_screen(xpos),
                       Config.AngleLineXColor)
        Render.rd_line(gui, world_to_screen(start), world_to_screen(ypos),
                       Config.AngleLineYColor)



from typing import List, Dict, Optional, Tuple

import numpy as np

from ...math.matrix import Matrix
from .aabb import AABB


class UniformGrid():
    def __init__(self):
        self._grid_size: float = 1.0

    def generate(self) -> List[Tuple[Matrix, Matrix]]:
        pass

    def raycast(self, p: Matrix, d: Matrix) -> Matrix:
        pass

    def update(self, body):
        pass

    def insert(self, body):
        pass

    def remove(self, body):
        pass



from abc import ABC, abstractmethod


class Frame(ABC):
    @abstractmethod
    def load(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def render(self) -> None:
        raise NotImplementedError



import copy

from functools import cmp_to_key
from typing import List, Tuple, Optional

import numpy as np

from ..math.matrix import Matrix


class GeomAlgo2D():
    class Clipper():
        @staticmethod
        def sutherland_hodgment_polygon_clipping(
                polygon: List[Matrix],
                clip_region: List[Matrix]) -> List[Matrix]:
            '''Sutherland Hodgman Polygon Clipping
            All points is stored in counter clock winding.
            By convention:
               p0 -> idx1 -> idx2 -> p0 constructs a triangle

            Parameters
            ----------
            polygon : List[Matrix]
                original polygon
            clip_region : List[Matrix]
                ref clip polygon

            Returns
            -------
            List[Matrix]
                clipped polygon
            '''
            res: List[Matrix] = copy.deepcopy(polygon)
            clip_len: int = len(clip_region)
            for i in range(clip_len - 1):
                clipa: Matrix = clip_region[i]
                clipb: Matrix = clip_region[i + 1]
                clipdir: Matrix = clip_region[
                    1] if i + 2 == clip_len else clip_region[i + 2]

                is_same_side: List[bool] = []
                poly_len: int = len(res)
                for j in range(poly_len):
                    is_same_side.append(
                        GeomAlgo2D.is_point_on_same_side(
                            clipa, clipb, clipdir, res[j]))

                new_polygon: List[Matrix] = []
                for j in range(1, poly_len):
                    last_inside: bool = is_same_side[j - 1]
                    cur_inside: bool = is_same_side[j]

                    # last inside and cur outside
                    if last_inside and not cur_inside:
                        # push last point
                        new_polygon.append(res[j - 1])
                        # push intersection point
                        new_polygon.append(
                            GeomAlgo2D.line_intersection(
                                clipa, clipb, res[j - 1], res[j]))
                    elif not last_inside and cur_inside:
                        # push intersection point first
                        new_polygon.append(
                            GeomAlgo2D.line_intersection(
                                clipa, clipb, res[j - 1], res[j]))
                    elif not last_inside and not cur_inside:
                        pass
                    elif last_inside and cur_inside:
                        new_polygon.append(res[j - 1])

                res = copy.deepcopy(new_polygon)
                res.append(res[0])

            return res

    @staticmethod
    def is_collinear(pa: Matrix, pb: Matrix, pc: Matrix) -> bool:
        '''check if point pa, pb, pc are collinear

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        bool
            True: collinear False: not collinear
        '''
        return np.isclose((pa - pb).cross(pa - pc), [0.0, 0.0]).all()

    @staticmethod
    def judge_range(val: float, low: float, high: float) -> bool:
        return low <= val <= high

    @staticmethod
    def is_fuzzy_collinear(pa: Matrix, pb: Matrix, target: Matrix) -> bool:
        '''check if the target point is in rectangle which diag is pa-pb

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        target : Matrix
            target point

        Returns
        -------
        bool
            True: in range False: no in range
        '''
        x_min = np.fmin(pa.x, pb.x)
        x_max = np.fmax(pa.x, pb.x)
        y_min = np.fmin(pa.y, pb.y)
        y_max = np.fmax(pa.y, pb.y)
        return GeomAlgo2D.judge_range(target.x, x_min,
                                      x_max) and GeomAlgo2D.judge_range(
                                          target.y, y_min, y_max)

    @staticmethod
    def is_point_on_segment(pa: Matrix, pb: Matrix, target: Matrix) -> bool:
        '''check if the target point is on line segment pa-pb

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        target : Matrix
            target point

        Returns
        -------
        bool
            True: on segment False: not on
        '''
        return GeomAlgo2D.is_collinear(
            pa, pb, target) and GeomAlgo2D.is_fuzzy_collinear(pa, pb, target)

    @staticmethod
    def line_segment_intersection(pa: Matrix, pb: Matrix, pc: Matrix,
                                  pd: Matrix) -> Optional[Matrix]:
        '''calculate intersected point between line pa-pb and line pc-pd.
        Return if there is a actual intersected point.
        Notices: overlapping is NOT considered as a kind of intersection
        situation in this function

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c
        pd : Matrix
            point d

        Returns
        -------
        Optional[Matrix]
            If exist, return intersected point, otherwise None
        '''

        ab: Matrix = pb - pa
        cd: Matrix = pd - pc
        ac: Matrix = pc - pa
        ad: Matrix = pd - pa
        bc: Matrix = pc - pb

        d1: float = Matrix.cross_product(ab, ac)
        d2: float = Matrix.cross_product(ab, ad)
        d3: float = Matrix.cross_product(cd, -ac)
        d4: float = Matrix.cross_product(cd, -bc)

        # XXX: need to simplify the code by abstract calc
        # intersection point method with nxt methods
        def _calc_pos(ab: Matrix, ac: Matrix, cd: Matrix,
                      pc: Matrix) -> Matrix:
            t: float = Matrix.cross_product(ab, ac) / Matrix.cross_product(
                cd, ab)
            return Matrix([pc.x + t * cd.x, pc.y + t * cd.y], 'vec')

        if d1 * d2 < 0 and d3 * d4 < 0:
            return _calc_pos(ab, ac, cd, pc)
        elif np.isclose(d1, 0) and GeomAlgo2D.is_point_on_segment(pa, pb, pc):
            return _calc_pos(ab, ac, cd, pc)
        elif np.isclose(d2, 0) and GeomAlgo2D.is_point_on_segment(pa, pb, pd):
            return _calc_pos(ab, ac, cd, pc)
        elif np.isclose(d3, 0) and GeomAlgo2D.is_point_on_segment(pc, pd, pa):
            return _calc_pos(ab, ac, cd, pc)
        elif np.isclose(d4, 0) and GeomAlgo2D.is_point_on_segment(pc, pd, pb):
            return _calc_pos(ab, ac, cd, pc)
        else:
            return None

    @staticmethod
    def line_intersection(pa: Matrix, pb: Matrix, pc: Matrix,
                          pd: Matrix) -> Matrix:
        '''check if line pa-pb and pc-pd is intersection
        if yes, return the intersection point
        https://blog.csdn.net/qq_45735851/article/details/114434281

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c
        pd : Matrix
            point d

        Returns
        -------
        Optional[Matrix]
            if insect, return insect point, otherwise return None
        '''
        linea: Matrix = pb - pa  # w
        lineb: Matrix = pd - pc  # v
        if np.isclose(Matrix.cross_product(linea, lineb), 0):
            return Matrix([0.0, 0.0], 'vec')

        u: Matrix = pc - pa
        t: float = Matrix.cross_product(linea, u) / Matrix.cross_product(
            lineb, linea)
        res: Matrix = Matrix([pc.x + t * lineb.x, pc.y + t * lineb.y], 'vec')

        return res

    @staticmethod
    def triangle_circum_center(pa: Matrix, pb: Matrix,
                               pc: Matrix) -> Optional[Matrix]:
        '''calc the circum-circle center from triangle pa-pb-pc

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        Optional[Matrix]
            If pa-pb-pc form triangle, return circum-circle center,
            otherwise None
        '''
        if np.isclose(GeomAlgo2D.triangle_area(pa, pb, pc), 0):
            return None

        # 2 * (x2 - x1) * x + 2 * (y2 - y1) y
        # = x2 ^ 2 + y2 ^ 2 - x1 ^ 2 - y1 ^ 2
        # 2 * (x3 - x2) * x + 2 * (y3 - y2) y
        # = x3 ^ 2 + y3 ^ 2 - x2 ^ 2 - y2 ^ 2
        val11: float = 2.0 * (pb.x - pa.x)
        val12: float = 2.0 * (pb.y - pa.y)
        val21: float = 2.0 * (pc.x - pb.x)
        val22: float = 2.0 * (pc.y - pb.y)
        coef_mat: Matrix = Matrix([val11, val12, val21, val22])
        equal_val: Matrix = Matrix([
            pb.len_square() - pa.len_square(),
            pc.len_square() - pb.len_square()
        ], 'vec')

        return coef_mat.invert() * equal_val

    @staticmethod
    def triangle_inscribed_center(pa: Matrix, pb: Matrix,
                                  pc: Matrix) -> Optional[Matrix]:
        '''calc the inscribed-circle center from triangle pa-pb-pc

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        Optional[Matrix]
            If pa-pb-pc form triangle, return inscribed-circle center,
            otherwise None
        '''
        if np.isclose(GeomAlgo2D.triangle_area(pa, pb, pc), 0):
            return None

        ab_len: float = (pb - pa).len()
        bc_len: float = (pc - pb).len()
        ca_len: float = (pc - pa).len()
        return (pc * ab_len + pa * bc_len + pb * ca_len) / (ab_len + bc_len +
                                                            ca_len)

    @staticmethod
    def calc_circum_center(pa: Matrix, pb: Matrix,
                           pc: Matrix) -> Optional[Tuple[Matrix, float]]:
        '''return the circum-circle's center and radius from triangle pa-pb-pc

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        Optional[Tuple[Matrix, float]]
            If can, return the (center, radius), otherwise None
        '''
        if np.isclose(GeomAlgo2D.triangle_area(pa, pb, pc), 0):
            return None

        center: Optional[Matrix] = GeomAlgo2D.triangle_circum_center(
            pa, pb, pc)
        assert center is not None
        radius: float = (center - pa).len()
        return (center, radius)

    @staticmethod
    def calc_inscribed_center(pa: Matrix, pb: Matrix,
                              pc: Matrix) -> Optional[Tuple[Matrix, float]]:
        '''return the inscribed-circle's center and radius
        from triangle pa-pb-pc

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        Optional[Tuple[Matrix, float]]
            If can, return the (center, radius), otherwise None
        '''
        area: float = GeomAlgo2D.triangle_area(pa, pb, pc)
        if np.isclose(area, 0):
            return None

        center: Optional[Matrix] = GeomAlgo2D.triangle_inscribed_center(
            pa, pb, pc)
        assert center is not None

        ab_len: float = (pb - pa).len()
        bc_len: float = (pc - pb).len()
        ca_len: float = (pc - pa).len()
        radius: float = 2.0 * area / (ab_len + bc_len + ca_len)

        return (center, radius)

    @staticmethod
    def is_convex_polygon(vertices: List[Matrix]) -> bool:
        '''check if the polygon is convex

        Parameters
        ----------
        vertices : List[Matrix]
            polygon's points

        Returns
        -------
        bool
            True: is convex, False: not is
        '''

        vert_len: int = len(vertices)
        if vert_len == 4:
            return True

        for i in range(vert_len - 1):
            ab: Matrix = vertices[i + 1] - vertices[i]
            ac: Matrix = vertices[i + 2] - vertices[
                i] if i + 2 != vert_len else vertices[1] - vertices[i]

            if Matrix.cross_product(ab, ac) < 0:
                return False

        return True

    @staticmethod
    def graham_scan_cmp(a: Matrix, b: Matrix) -> int:
        first_val1: float = np.arctan2(a.y, a.x)
        first_val2: float = np.arctan2(b.y, b.x)
        if not np.isclose(first_val1, first_val2):
            if first_val1 < first_val2:
                return -1
            elif first_val1 > first_val2:
                return 1
        else:
            second_val1: float = a.x
            second_val2: float = b.x
            if not np.isclose(second_val1, second_val2):
                if second_val1 < second_val2:
                    return -1
                elif second_val1 > second_val2:
                    return 1
        return 0

    @staticmethod
    def graham_scan(vertices: List[Matrix]) -> List[Matrix]:
        '''convex hull algorithm: Graham Scan. Given a series of points,
        find the convex polygon that can contains all of these points.

        Parameters
        ----------
        vertices : List[Matrix]
            polygon's points

        Returns
        -------
        List[Matrix]
            convex polygon's point
        '''
        sort_vert: List[Matrix] = copy.deepcopy(vertices)
        sort_vert.sort(key=cmp_to_key(GeomAlgo2D.graham_scan_cmp))

        stack: List[int] = []
        check_idx: int = 2
        stack.append(0)
        stack.append(1)
        while True:
            i: int = stack[-2]
            j: int = stack[-1]
            if j == 0:
                break

            if check_idx >= len(sort_vert):
                check_idx = 0

            ab: Matrix = sort_vert[j] - sort_vert[i]
            ac: Matrix = sort_vert[check_idx] - sort_vert[i]

            if ab.cross(ac) < 0:
                stack.pop()
            stack.append(check_idx)
            check_idx += 1

        res: List[Matrix] = []
        for idx in stack:
            res.append(sort_vert[idx])

        return res

    @staticmethod
    def point_to_line_segment(pa: Matrix, pb: Matrix, pc: Matrix) -> Matrix:
        '''calculate point on line segment pa-pb that is
        the shortest length to pc

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        Matrix
            shortest-dis point to the segment, otherwise zero
        '''

        if pa == pb:
            return Matrix([0.0, 0.0], 'vec')

        if GeomAlgo2D.is_collinear(pa, pb, pc):
            if GeomAlgo2D.is_point_on_segment(pa, pb, pc):
                return pc
            else:
                return pb if (pc - pa).len_square() > (pc -
                                                       pb).len_square() else pa

        ac: Matrix = pc - pa
        ab_normal: Matrix = (pb - pa).normal()
        ac_proj: Matrix = ab_normal * ab_normal.dot(ac)
        op_proj: Matrix = pa + ac_proj

        if GeomAlgo2D.is_fuzzy_collinear(pa, pb, op_proj):
            return op_proj

        return pb if (pc - pa).len_square() > (pc - pb).len_square() else pa

    @staticmethod
    def shortest_length_point_of_ellipse(a: float, b: float,
                                         pc: Matrix) -> Matrix:
        '''calculate point on ellipse that is the shortest length
        to pc(aka projection point)

        Parameters
        ----------
        a : float
            semi-major axis len
        b : float
            semi-minor axis len
        pc : Matrix
            target point

        Returns
        -------
        Optional[Matrix]
            shortest-dis point to the ellipse, otherwise None
        '''

        if np.isclose(a, 0) or np.isclose(b, 0):
            return Matrix([0.0, 0.0], 'vec')

        if np.isclose(pc.x, 0):
            return Matrix([0.0, b], 'vec') if pc.y > 0 else Matrix([0.0, -b],
                                                                   'vec')

        if np.isclose(pc.y, 0):
            return Matrix([a, 0.0], 'vec') if pc.x > 0 else Matrix([-a, 0.0],
                                                                   'vec')

        x_left: float = 0.0
        x_right: float = 0.0
        t0: Matrix = Matrix([0.0, 0.0], 'vec')
        t1: Matrix = Matrix([0.0, 0.0], 'vec')
        sgn: int = 1 if pc.y > 0 else -1

        if pc.x < 0:
            x_left = -a
            x_right = 0.0
        else:
            x_left = 0.0
            x_right = a

        iter_val: int = 0
        while True:
            iter_val += 1

            tmp_x0: float = (x_left + x_right) * 0.5
            tmp_y0: float = sgn * np.sqrt(b**2 - (b / a)**2 * tmp_x0**2)
            t0.set_value([tmp_x0, tmp_y0])

            tmp_x1: float = tmp_x0 + 1
            tmp_y1: float = (b**2 - (b / a)**2 * tmp_x1 * tmp_x0) / tmp_y0
            t1.set_value([tmp_x1, tmp_y1])

            t0t1: Matrix = t1 - t0
            t0p: Matrix = pc - t0

            res: float = t0t1.dot(t0p)
            if np.isclose(np.fabs(res), 0.0):
                break

            if res > 0:
                x_left = tmp_x0
            else:
                x_right = tmp_x0

        return t0

    @staticmethod
    def triangle_centroid(pa: Matrix, pb: Matrix, pc: Matrix) -> Matrix:
        '''calculate the centroid of the triangle

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        Matrix
            centroid point
        '''
        return (pa + pb + pc) / 3.0

    @staticmethod
    def triangle_area(pa: Matrix, pb: Matrix, pc: Matrix) -> float:
        '''calculate the area of the triangle

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        float
            triangle area
        '''
        return np.fabs(Matrix.cross_product(pa - pb, pa - pc) / 2.0)

    # NOTE: need to focus on the vertices format
    # the vertices format is for the close shape
    @staticmethod
    def calc_mass_center(vertices: List[Matrix]) -> Matrix:
        '''calculate the mass center of the convex polygon

        Parameters
        ----------
        vertices : List[Matrix]
            polygon's points

        Returns
        -------
        Matrix
            mass center
        '''

        vert_len: int = len(vertices)
        if vert_len >= 4:
            pos: Matrix = Matrix([0.0, 0.0], 'vec')
            tot_area: float = 0.0

            for i in range(vert_len - 1):
                idx1: int = i + 1
                idx2: int = i + 2
                if idx1 == vert_len - 2:
                    break

                tri_area: float = GeomAlgo2D.triangle_area(
                    vertices[0], vertices[idx1], vertices[idx2])
                tri_centroid: Matrix = GeomAlgo2D.triangle_centroid(
                    vertices[0], vertices[idx1], vertices[idx2])

                pos += tri_centroid * tri_area
                tot_area += tri_area

            pos /= tot_area

            return pos
        else:
            return Matrix([0.0, 0.0], 'vec')

    @staticmethod  # FIXME: need to understand the algorithm
    def shortest_length_line_segment_ellipse(
            a: float, b: float, pc: Matrix,
            pd: Matrix) -> Tuple[Matrix, Matrix]:
        '''calculate two points on line segment and ellipse respectively.
        The length of two points is the shortest distance of line segment
        and ellipse

        Parameters
        ----------
        a : float
            semi-major axis len
        b : float
            semi-minor axis len
        pc : Matrix
            line segment point c
        pd : Matrix
            line segment point d

        Returns
        -------
        Tuple[Matrix, Matrix]
            the return data
        '''

        res_segm: Matrix = Matrix([0.0, 0.0], 'vec')
        res_elli: Matrix = Matrix([0.0, 0.0], 'vec')
        tmp: float = 0.0
        # FIXME: can write helper func
        if np.isclose(pc.y, pd.y):
            if not ((pc.x > 0 and pd.x > 0) or (pc.x < 0 and pd.x < 0)):
                res_elli.set_value([0.0, b if pc.y > 0 else -b])
                res_segm.set_value([0.0, pc.y])
            else:
                tmp = np.fmin(np.fabs(pc.x), np.fabs(pd.x))
                res_segm.set_value([tmp, pc.y])
                res_elli = GeomAlgo2D.shortest_length_point_of_ellipse(
                    a, b, res_segm)
        elif np.isclose(pc.x, pd.x):
            if not ((pc.y > 0 and pd.y > 0) or (pc.y < 0 and pd.y < 0)):
                res_elli.set_value([a if pc.x > 0 else -a, 0.0])
                res_segm.set_value([pc.x, 0.0])
            else:
                tmp = np.fmin(np.fabs(pc.y), np.fabs(pd.y))
                res_segm.set_value([pc.x, tmp])
                res_elli = GeomAlgo2D.shortest_length_point_of_ellipse(
                    a, b, res_segm)
        else:
            k: float = (pd.y - pc.y) / (pd.x - pc.x)
            k2: float = k * k
            a2: float = a**2
            b2: float = b**2
            f_x2: float = (k2 * a2 * a2 / b2) / (1 + a2 * k2 / b2)
            f_y2: float = b2 - b2 * f_x2 / a2
            f_x: float = np.sqrt(f_x2)
            f_y: float = np.sqrt(f_y2)
            f: Matrix = Matrix([0.0, 0.0], 'vec')
            pcpd: Matrix = (pd - pc).normal()

            # Check which quadrant does nearest point fall in
            f_arr: List[Matrix] = []
            f_arr.append(Matrix([f_x, f_y], 'vec'))
            f_arr.append(Matrix([-f_x, f_y], 'vec'))
            f_arr.append(Matrix([-f_x, -f_y], 'vec'))
            f_arr.append(Matrix([f_x, -f_y], 'vec'))

            min_val: float = Matrix.cross_product(pcpd, f_arr[0] - pc)
            for i in range(1, 4):
                tmp_val: float = Matrix.cross_product(pcpd, f_arr[i] - pc)
                if tmp_val < min_val:
                    f = f_arr[i]
                    min_val = tmp_val

            pcf: Matrix = f - pc
            pc_fp: Matrix = pcpd * pcpd.dot(pcf)
            f_proj: Matrix = pc + pc_fp

            # NOTE: focus the projection dirn
            if GeomAlgo2D.is_fuzzy_collinear(Matrix([a, 0.0], 'vec'),
                                             Matrix([b, 0.0], 'vec'), f_proj):
                res_elli = f
                res_segm = f_proj
            else:
                pc_p = GeomAlgo2D.shortest_length_point_of_ellipse(a, b, pc)
                pd_p = GeomAlgo2D.shortest_length_point_of_ellipse(a, b, pd)

                if (pc - pc_p).len_square() > (pd - pd_p).len_square():
                    res_elli = pd_p
                    res_segm = pd
                else:
                    res_elli = pc_p
                    res_segm = pc

        return (res_elli, res_segm)

    @staticmethod
    def raycast(pa: Matrix, dirn: Matrix, pc: Matrix,
                pd: Matrix) -> Optional[Matrix]:
        '''calculate point on line segment pc-pd, if point 'pa'
        can cast ray in 'dirn' direction on line segment pc-pd.

        Parameters
        ----------
        pa : Matrix
            ray start point
        dirn : Matrix
            ray direction
        pc : Matrix
            line segment point a
        pd : Matrix
            line segment point b

        Returns
        -------
        Optional[Matrix]
            If exist, return raycast point, otherwise None
        '''

        denominator: float = (pa.x - dirn.x) * (pc.y - pd.y) - (
            pa.y - dirn.y) * (pc.x - pd.x)

        if np.isclose(denominator, 0):
            return None

        t: float = ((pa.x - pc.x) * (pc.y - pd.y) - (pa.y - pc.y) *
                    (pc.x - pd.x)) / denominator
        u: float = ((dirn.x - pa.x) * (pa.y - pc.y) - (dirn.y - pa.y) *
                    (pa.x - pc.x)) / denominator

        if t >= 0 and 0 <= u <= 1.0:
            return Matrix(
                [pa.x + t * (dirn.x - pa.x), pa.y + t * (dirn.y - pa.y)],
                'vec')

        return None

    @staticmethod
    def raycast_aabb(pa: Matrix, dirn: Matrix, top_left: Matrix,
                     bot_right: Matrix) -> Optional[Tuple[Matrix, Matrix]]:
        '''calculate point on  AABB, if point 'pa' can cast ray in 'dirn'
        direction on AABB.

        Parameters
        ----------
        pa : Matrix
            ray start point
        dirn : Matrix
            ray direction
        top_left : Matrix
            AABB top left point
        bot_right : Matrix
            AABB bot right point

        Returns
        -------
        Optional[Matrix]
            If exist, return two raycast point, one is incidence point, another
        is emission point, otherwise None
        '''

        x_min: float = top_left.x
        y_min: float = bot_right.y
        x_max: float = bot_right.x
        y_max: float = top_left.y
        tx_min: float = 0.0
        ty_min: float = 0.0
        tx_max: float = 0.0
        ty_max: float = 0.0
        tx_enter: float = 0.0
        tx_exit: float = 0.0
        ty_enter: float = 0.0
        ty_exit: float = 0.0
        t_enter: float = 0.0
        t_exit: float = 0.0

        if np.isclose(dirn.x, 0) and not np.isclose(dirn.y, 0):
            ty_min = (y_min - pa.y) / dirn.y
            ty_max = (y_max - pa.y) / dirn.y
            t_enter = np.fmin(ty_min, ty_max)
            t_exit = np.fmax(ty_min, ty_max)
        elif not np.isclose(dirn.x, 0) and np.isclose(dirn.y, 0):
            tx_min = (x_min - pa.x) / dirn.x
            tx_max = (x_max - pa.x) / dirn.x
            t_enter = np.fmin(tx_min, tx_max)
            t_exit = np.fmax(tx_min, tx_max)
        else:
            tx_min = (x_min - pa.x) / dirn.x
            tx_max = (x_max - pa.x) / dirn.x
            ty_min = (y_min - pa.y) / dirn.y
            ty_max = (y_max - pa.y) / dirn.y
            tx_enter = np.fmin(tx_min, tx_max)
            tx_exit = np.fmax(tx_min, tx_max)
            ty_enter = np.fmin(ty_min, ty_max)
            ty_exit = np.fmax(ty_min, ty_max)
            t_enter = np.fmax(tx_enter, ty_enter)
            t_exit = np.fmin(tx_exit, ty_exit)

        if t_enter < 0 and t_exit < 0:
            return None

        # NOTE: just for different with buildin type
        out_enter: Matrix = pa + dirn * t_enter
        out_exit: Matrix = pa + dirn * t_exit

        return (out_enter, out_exit)

    @staticmethod
    def is_point_on_AABB(pa: Matrix, top_left: Matrix,
                         bot_right: Matrix) -> bool:
        '''check if the pa is in the AABB

        Parameters
        ----------
        pa : Matrix
            target point
        top_left : Matrix
            AABB top left point
        bot_right : Matrix
            AABB bottom right point

        Returns
        -------
        bool
            True: is in AABB False: is not
        '''

        return GeomAlgo2D.judge_range(pa.x, top_left.x,
                                      bot_right.x) and GeomAlgo2D.judge_range(
                                          pa.y, bot_right.y, top_left.y)

    @staticmethod
    def rotate(pa: Matrix, center: Matrix, radian: float) -> Matrix:
        '''rotate point 'pa' around point 'center' by radian

        Parameters
        ----------
        pa : Matrix
            source point
        center : Matrix
            center point
        radian : float
            rotate radian

        Returns
        -------
        Matrix
            result point
        '''

        return Matrix.rotate_mat(radian) * (pa - center) + center

    @staticmethod
    def calc_ellipse_project_on_point(a: float, b: float,
                                      dirn: Matrix) -> Matrix:
        '''calculate the projection axis of ellipse in user-define direction.
        return the maximum point in ellipse

        Parameters
        ----------
        a : float
            semi-major axis len
        b : float
            semi-minor axis len
        dirn : Matrix
            user define direction

        Returns
        -------
        Matrix
            maximum point
        '''

        res: Matrix = Matrix([0.0, 0.0], 'vec')
        sgn: int = -1

        if np.isclose(dirn.x, 0):
            sgn = -1 if dirn.y < 0 else 1
            res.set_value([0.0, sgn * b])
        elif np.isclose(dirn.y, 0):
            sgn = -1 if dirn.x < 0 else 1
            res.set_value([sgn * a, 0.0])
        else:
            # y = kx
            k: float = dirn.y / dirn.x
            # line offset constant
            a2: float = a**2
            b2: float = b**2
            k2: float = k**2
            d: float = np.sqrt((a2 + b2 * k2) / k2)
            if Matrix.dot_product(Matrix([0.0, d], 'vec'), dirn) < 0:
                d = d * -1

            x1: float = k * d - (b2 * k2 * k * d) / (a2 + b2 * k2)
            y1: float = (b2 * k2 * d) / (a2 + b2 * k2)
            res.set_value([x1, y1])

        return res

    @staticmethod
    def calc_capsule_project_on_point(width: float, height: float,
                                      dirn: Matrix) -> Matrix:
        '''calculate the projection axis of capsule in user-define direction.
        return the maximum point in capsule

        Parameters
        ----------
        width : float
            capsule width
        height : float
            capsule height
        dirn : Matrix
            user define direction

        Returns
        -------
        Matrix
            maximum point
        '''

        res: Matrix = Matrix([0.0, 0.0], 'vec')
        radius: float = 0.0
        offset: float = 0.0
        if width > height:
            radius = height / 2.0
            if dirn.x >= 0:
                offset = width / 2.0 - radius
            else:
                offset = radius - width / 2.0

            res = dirn.normal() * radius
            res.x += offset
        else:
            radius = width / 2.0
            if dirn.y >= 0:
                offset = height / 2.0 - radius
            else:
                offset = radius - height / 2.0

            res = dirn.normal() * radius
            res.y += offset

        return res

    @staticmethod
    def calc_sector_project_on_point(start: float, span: float, radius: float,
                                     dirn: Matrix) -> Matrix:
        '''calculate the projection axis of sector in user-define direction.
        return the maximum point in sector

        Parameters
        ----------
        start : float
            start radian
        span : float
            span radian(delta radian)
        radius : float
            radius value
        dirn : Matrix
            user define direction

        Returns
        -------
        Matrix
            maximum point
        '''

        res: Matrix = Matrix([0.0, 0.0], 'vec')

        def _clamp_radian(radian: float) -> float:
            _res: float = radian
            _res -= np.floor(_res / np.pi * 2) * np.pi * 2

            if _res < 0:
                _res += np.pi * 2

            return _res

        clamp_start: float = _clamp_radian(start)
        clamp_end: float = _clamp_radian(start + span)
        origin_start: float = _clamp_radian(start - np.pi / 2.0)
        origin_end: float = _clamp_radian(start + span + np.pi / 2.0)
        origin_theta: float = dirn.theta()
        theta: float = _clamp_radian(origin_theta)

        if origin_start > origin_end:
            # does not fall in zero area
            if not GeomAlgo2D.judge_range(theta, origin_end, origin_start):
                if theta > origin_start:
                    theta = origin_theta

                # clamp theta to sector area
                res = Matrix.rotate_mat(np.clip(
                    theta, clamp_start, clamp_end)) * Matrix([1.0, 0.0],
                                                             'vec') * radius

        elif origin_start < origin_end:
            if GeomAlgo2D.judge_range(origin_theta, origin_start, origin_end):
                res = Matrix.rotate_mat(np.clip(
                    theta, clamp_start, clamp_end)) * Matrix([1.0, 0.0],
                                                             'vec') * radius

        if np.isclose(origin_start, origin_end):
            if not np.isclose(theta, origin_start):
                if theta > origin_start:
                    theta = origin_theta

                if clamp_start > clamp_end:
                    res = Matrix.rotate_mat(
                        np.clip(theta, clamp_start - np.pi * 2.0,
                                clamp_end)) * Matrix([1.0, 0.0],
                                                     'vec') * radius
                else:
                    res = Matrix.rotate_mat(
                        np.clip(theta, clamp_start, clamp_end)) * Matrix(
                            [1.0, 0.0], 'vec') * radius

        return res

    @staticmethod
    def is_triangle_contain_origin(pa: Matrix, pb: Matrix, pc: Matrix) -> bool:
        '''check if the origin point is in the triangle

        Parameters
        ----------
        pa : Matrix
            point a
        pb : Matrix
            point b
        pc : Matrix
            point c

        Returns
        -------
        bool
            True: origin point is in triangle, otherwise not
        '''

        ra: float = (pb - pa).cross(-pa)
        rb: float = (pc - pb).cross(-pb)
        rc: float = (pa - pc).cross(-pc)
        return (ra >= 0 and rb >= 0 and rc >= 0) or (ra <= 0 and rb <= 0
                                                     and rc <= 0)

    @staticmethod
    def is_point_on_same_side(edgp1: Matrix, edgp2: Matrix, ref: Matrix,
                              target: Matrix) -> bool:
        '''check if the target and ref point is on edgep1-edgep2 same side

        Parameters
        ----------
        edgp1 : Matrix
            edge point 1
        edgp2 : Matrix
            edge point 2
        ref : Matrix
            ref point
        target : Matrix
            target point

        Returns
        -------
        bool
            True: is on same side, otherwise not
        '''

        u = edgp1 - edgp2
        v = ref - edgp1
        w = target - edgp1
        d1 = u.cross(v)
        d2 = u.cross(w)
        return np.sign(d1) == np.sign(d2)



from typing import List, Dict, Optional, Tuple

import numpy as np

from ..broad_phase.aabb import AABB
from ...dynamics.body import Body
from ..detector import Detector
from ..broad_phase.dbvh import DBVH
from ...math.matrix import Matrix
from ...common.config import Config


class CCD():
    '''Continuous Collision Detection

    This class is implemented by bisection and re-sampling. Both them are costly.
    '''
    class AABBShot():
        def __init__(self,
                     box: AABB,
                     attr: Body.PhysicsAttribute,
                     t: float = 0.0):
            self._aabb: AABB = box
            self._attr: Body.PhysicsAttribute = attr
            self._time: float = t

    class IndexSection():
        def __init__(self):
            self._forward: int = -1
            self._backward: int = -1

    class CCDPair():
        def __init__(self, time: float = 0.0, target: Body = None):
            self._toi: float = time
            self._body: Optional[Body] = target

    @staticmethod
    def build_trajectory_aabb(
            body: Body,
            dt: float,
            target: Optional[Matrix] = None) -> Tuple[List[AABBShot], AABB]:
        assert body != None

        traj = []  # HACK: add type hint
        res: AABB = AABB()

        start: Body.PhysicsAttribute = body.phy_attr
        start_box: AABB = AABB.from_body(body)

        # run a dt step
        body.step_position(dt)

        # end: Body.PhysicsAttribute= body.phy_attr
        end_box: AABB = AABB.from_body(body)

        if start_box == end_box and start._vel.len_square(
        ) < Config.MaxVelocity and np.fabs(
                start._ang_vel) < Config.MaxAngularVelocity:
            traj.append(CCD.AABBShot(start_box, body.phy_attr, 0))
            traj.append(CCD.AABBShot(end_box, body.phy_attr, dt))
            return (traj, res)

        res.unite(start_box).unite(end_box)
        body.phy_attr = start

        slice: float = 40.0
        step: float = dt / slice
        idx: float = step
        while idx <= dt:
            body.step_position(step)
            aabb: AABB = AABB.from_body(body)
            traj.append(CCD.AABBShot(aabb, body.phy_attr, idx))
            res.unite(aabb)
            idx += step
            if target != None and (aabb.pos - target).len_square() >= (
                    target - start._pos).len_square():
                break

        body.phy_attr = start
        return (traj, res)

    @staticmethod
    def find_broad_phase_root(bodya: Body, traja: List[AABBShot], bodyb: Body,
                              trajb: List[AABBShot],
                              dt: float) -> Optional[IndexSection]:
        assert bodya != None and bodyb != None
        is_bodya_ccd: bool = len(traja) > 2
        is_bodyb_ccd: bool = len(trajb) > 2

        if is_bodya_ccd and is_bodyb_ccd:
            traj1: AABB = AABB.unite(traja[0]._aabb,
                                     traja[len(traja) - 1]._aabb)
            traj2: AABB = AABB.unite(trajb[0]._aabb,
                                     trajb[len(trajb) - 1]._aabb)

            if not traj1.collide(traj2):
                return None

            length: int = np.fmin(len(traja), len(trajb))
            res: CCD.IndexSection = CCD.IndexSection()

            for i in range(length - 1):
                tmp1: AABB = AABB.unite(traja[i]._aabb, traja[i + 1]._aabb)
                tmp2: AABB = AABB.unite(trajb[i]._aabb, trajb[i + 1]._aabb)
                if tmp1.collide(tmp2):
                    res._forward = i
                    break

            return None if res._forward == -1 else res

        elif not is_bodya_ccd or not is_bodyb_ccd:
            traj_static: Optional[List[CCD.AABBShot]] = None
            traj_dynamic: Optional[List[CCD.AABBShot]] = None

            if is_bodya_ccd:
                traj_static = trajb
                traj_dynamic = traja
            elif is_bodyb_ccd:
                traj_static = traja
                traj_dynamic = trajb

            traj1: AABB = AABB.unite(
                traj_static.at(0)._aabb,
                traj_static.at(1)._aabb)
            traj2: AABB = AABB.unite(
                traj_dynamic.at(0)._aabb,
                traj_dynamic.at(len(traj_dynamic) - 1)._aabb)

            if not traj1.collide(traj2):
                return None

            res: CCD.IndexSection = CCD.IndexSection()
            length: int = len(traj_dynamic)
            forward_found: bool = False
            backward_found: bool = False

            j: int = length - 1
            for i in range(length) - 1:
                tmp_forward: AABB = AABB.unite(
                    traj_dynamic.at(i)._aabb,
                    traj_dynamic.at(i + 1)._aabb)

                tmp_backward: AABB = AABB.unite(
                    traj_dynamic.at(j)._aabb,
                    traj_dynamic.at(j - 1)._aabb)

                if tmp_forward.collide(traj1) and not forward_found:
                    res._forward = i
                    forward_found = True

                if tmp_backward.collide(traj1) and not backward_found:
                    res._backward = j
                    backward_found = True

                if forward_found and backward_found:
                    break

                j -= 1

            return None if res._forward == -1 else res

        return None

    @staticmethod
    def find_narrow_phase_root(bodya: Body, traja: List[AABBShot], bodyb: Body,
                               trajb: List[AABBShot], idx: IndexSection,
                               dt: float) -> Optional[float]:
        assert bodya != None and bodyb != None

        is_bodya_ccd: bool = len(traja) > 2
        is_bodyb_ccd: bool = len(trajb) > 2
        if is_bodya_ccd and is_bodyb_ccd:
            return None

        elif not is_bodya_ccd or not is_bodyb_ccd:
            dynamic_body: Body = None
            origin1: Body.PhysicsAttribute = bodya.phy_attr
            origin2: Body.PhysicsAttribute = bodyb.phy_attr

            start_timestep: float = 0.0
            end_timestep: float = 0.0

            if is_bodya_ccd:
                dynamic_body = bodya
                bodya.phy_attr = traja[idx._forward]._attr
                start_timestep = traja[idx._forward]._time
                end_timestep = traja[idx._backward]._time

            elif is_bodyb_ccd:
                dynamic_body = bodyb
                bodyb.phy_attr = trajb[idx._forward]._attr
                start_timestep = trajb[idx._forward]._time
                end_timestep = trajb[idx._backward]._time

            # slice maybe 25~70. It depends on how thin the sticks you set
            slice: float = 30.0
            step: float = (end_timestep - start_timestep) / slice
            forward_steps: float = 0

            last_attr: Body.PhysicsAttribute = Body.PhysicsAttribute()
            is_find: bool = False

            while start_timestep + forward_steps <= end_timestep:
                last_attr = dynamic_body.phy_attr
                dynamic_body.step_position(step)
                forward_steps += step

                res: Optional[bool] = Detector.collide(bodya, bodyb)

                if res != None:
                    forward_steps -= step
                    dynamic_body.phy_attr = last_attr
                    is_find = True
                    break

            if not is_find:
                bodya.phy_attr = origin1
                bodyb.phy_attr = origin2
                return None

            # retracing
            step /= 2.0
            epsilon: float = 0.01

            while start_timestep + forward_steps <= end_timestep:
                last_attr = dynamic_body.phy_attr
                dynamic_body.step_position(step)
                forward_steps += step

                res = Detector.detect(bodya, bodyb)
                if res._is_colliding:
                    if np.fabs(res._penetration) < epsilon:
                        bodya.phy_attr = origin1
                        bodyb.phy_attr = origin2
                        return start_timestep + forward_steps

                    forward_steps -= step
                    dynamic_body.phy_attr = last_attr
                    step /= 2.0

        return None

    @staticmethod
    def query(root: DBVH.Node, body: Optional[Body],
              dt: float) -> Optional[List[CCDPair]]:
        assert root.is_root() and body != None

        query_list: List[CCD.CCDPair] = []
        potential: List[DBVH.Node] = []
        (traj_ccd, aabb_ccd) = CCD.build_trajectory_aabb(body, dt)

        DBVH.query_nodes(root, aabb_ccd, potential, body)

        for elem in potential:
            (traj_elem, aabb_elem) = CCD.build_trajectory_aabb(elem._body, dt)
            (new_ccd_tarj,
             new_aabb) = CCD.build_trajectory_aabb(body, dt, elem._body.pos)
            res: Optional[CCD.IndexSection] = CCD.find_broad_phase_root(
                body, new_ccd_tarj, elem._body, traj_elem, dt)

            if res != None:
                toi: Optional[float] = CCD.find_narrow_phase_root(
                    body, new_ccd_tarj, elem._body, traj_elem, res, dt)
                if toi != None:
                    query_list.append(CCD.CCDPair(toi, elem._body))

        return query_list if len(query_list) > 0 else None
